import numpy as np

import sxs
import scri
import spinsfast
import spherical_functions as sf
from spherical_functions import LM_index

from sxs.waveforms.alignment import align2d
from scri.asymptotic_bondi_data.bms_charges import charge_vector_from_aspect

import quaternion
from quaternion.calculus import indefinite_integral as integrate

from scipy.interpolate import CubicSpline


def MT_to_WM(h_mts, sxs_version=False, dataType=scri.h):
    """Convert a ModesTimeSeries object to a scri or a sxs WaveformModes object.

    Parameters
    ----------
    h_mts: ModesTimesSeries
        ModesTimeSeries object to be converted to WaveformModes object.
    sxs_version: bool, default is False
        If True (False), then return the sxs (scri) WaveformModes object. Default is False.
    dataType: int, default is 7 (scri.h)
        Data type of the WaveformModes object, e.g., scri.h or scri.hdot. Default is 7, i.e., scri.h
    """
    if not sxs_version:
        h = scri.WaveformModes(
            t=h_mts.t,
            data=np.array(h_mts)[:, LM_index(abs(h_mts.s), -abs(h_mts.s), 0) :],
            ell_min=abs(h_mts.s),
            ell_max=h_mts.ell_max,
            frameType=scri.Inertial,
            dataType=dataType,
        )
        h.r_is_scaled_out = True
        h.m_is_scaled_out = True
        return h
    else:
        h = sxs.WaveformModes(
            input_array=np.array(h_mts)[:, LM_index(abs(h_mts.s), -abs(h_mts.s), 0) :],
            time=h_mts.t,
            time_axis=0,
            modes_axis=1,
            ell_min=abs(h_mts.s),
            ell_max=h_mts.ell_max,
            spin_weight=h_mts.s,
        )
        return h


def WM_to_MT(h_wm):
    """Convert a WaveformModes object to a ModesTimeSeries object.

    Parameters
    ----------
    h_wm: WaveformModes
        WaveformModes object to be converted to ModesTimeSeries object.
    """
    h_mts = scri.ModesTimeSeries(
        sf.SWSH_modes.Modes(
            h_wm.data,
            spin_weight=h_wm.spin_weight,
            ell_min=h_wm.ell_min,
            ell_max=h_wm.ell_max,
            multiplication_truncator=max,
        ),
        time=h_wm.t,
    )
    return h_mts


def 𝔇(h, ell_max):
    """Differential operator 𝔇 acting on spin-weight s=0 function.

    This is equivalent to ð^{2}\bar{ð}^{2}.

    It is explicitly defined in Eqs (16) and (17) of https://doi.org/10.1063/1.532646.

    """
    h_with_operator = h.copy()
    for ell in range(0, ell_max + 1):
        if ell < 2:
            value = 0
        else:
            value = ((ell + 2) * (ell + 1) * (ell) * (ell - 1)) / 4.0
        h_with_operator[..., LM_index(ell, -ell, 0) : LM_index(ell, ell, 0) + 1] *= value

    return h_with_operator


def 𝔇inverse(h, ell_max):
    """Inverse of differential operator 𝔇 acting on spin-weight s=0 function."""
    h_with_operator = h.copy()
    for ell in range(0, ell_max + 1):
        if ell < 2:
            value = 0
        else:
            value = 4.0 / ((ell + 2) * (ell + 1) * (ell) * (ell - 1))
        h_with_operator[..., LM_index(ell, -ell, 0) : LM_index(ell, ell, 0) + 1] *= value

    return h_with_operator


def compute_bondi_rest_mass_and_conformal_factor(PsiM, ell_max):
    """Compute the Bondi rest mass and conformal factor K.

    Compute the Bondi rest mass and the conformal factor K from
    the Moreschi supermomentum to help obtain the Moreschi supermomentum
    in a supertranslated frame. These are defined in Eqs. (14) and (15) of
    https://doi.org/10.1063/1.532646.

    Parameters
    ----------
    PsiM: ModesTimeSeries
        Moreschi supermomentum computed from AsymptoticBondiData object.
    ell_max: int
        Maximum ell to use when converting data from SWSHs to data on the spherical grid.
    """
    # This comes from Eq. (1), note that charge_vector_from aspect divides by np.sqrt(4 * np.pi)
    P = -charge_vector_from_aspect(PsiM)

    r_vector_modes_on_Grid = []
    for L, M in [(0, 0), (1, -1), (1, 0), (1, +1)]:
        r_vector_modes = np.zeros((ell_max + 1) ** 2)
        r_vector_modes[LM_index(L, M, 0)] = 1
        r_vector_modes_on_Grid.append(spinsfast.salm2map(r_vector_modes, 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1))

    # This comes from Eq. (12), note that charge_vector_from aspect divides by np.sqrt(4 * np.pi)
    r_vector = 4 * np.pi * charge_vector_from_aspect(np.array(r_vector_modes_on_Grid).transpose()).transpose()

    # These come from Eqs. (14) and (15)
    if len(PsiM.shape) > 1:
        M_Grid = np.sqrt(P[:, 0] ** 2 - (P[:, 1] ** 2 + P[:, 2] ** 2 + P[:, 3] ** 2))

        K_Grid = M_Grid[:, None, None] / (
            np.tensordot(P[:, 0], r_vector[0], axes=0)
            - (
                np.tensordot(P[:, 1], r_vector[1], axes=0)
                + np.tensordot(P[:, 2], r_vector[2], axes=0)
                + np.tensordot(P[:, 3], r_vector[3], axes=0)
            )
        )
    else:
        M_Grid = np.sqrt(P[0] ** 2 - (P[1] ** 2 + P[2] ** 2 + P[3] ** 2))

        K_Grid = M_Grid / (P[0] * r_vector[0] - (P[1] * r_vector[1] + P[2] * r_vector[2] + P[3] * r_vector[3]))

    return M_Grid, K_Grid


def compute_Moreschi_supermomentum(PsiM, alpha, ell_max):
    """Compute the Moreschi supermomentum in a supertranslated frame.

    The transformation of the Moreschi supermomentum can be found in
    Eq (9) of https://doi.org/10.1063/1.532646.

    Parameters
    ----------
    PsiM: ModesTimeSeries
        Moreschi supermomentum computed from AsymptoticBondiData object.
    alpha: ndarray, complex, shape (Ntheta, Nphi)
        Supertranslation to apply to Moreschi supermomentum.
        This should be stored as an array of spherical harmonic modes.
    ell_max: int
        Maximum ell to use when converting data from SWSHs to data on the spherical grid.
    """
    M_Grid, K_Grid = compute_bondi_rest_mass_and_conformal_factor(np.array(PsiM), ell_max)

    PsiM_Grid = PsiM.grid().real
    PsiM_Grid_interp = sf.SWSH_grids.Grid(np.zeros((2 * ell_max + 1, 2 * ell_max + 1), dtype=float), spin_weight=0)
    K_Grid_interp = sf.SWSH_grids.Grid(np.zeros((2 * ell_max + 1, 2 * ell_max + 1), dtype=float), spin_weight=0)

    # interpolate these functions onto the new retarded time
    for i in range(2 * ell_max + 1):
        for j in range(2 * ell_max + 1):
            alpha_i_j = alpha[i, j]
            PsiM_Grid_interp[i, j] = CubicSpline(PsiM.t, PsiM_Grid[:, i, j])(alpha_i_j)

            K_Grid_interp[i, j] = CubicSpline(PsiM.t, K_Grid[:, i, j])(alpha_i_j)

    # compute the supertranslation term
    alpha_modes = spinsfast.map2salm(alpha.view(np.ndarray), 0, ell_max)
    D_alpha_modes = 𝔇(alpha_modes, ell_max)
    D_alpha_modes_Grid = sf.SWSH_grids.Grid(
        spinsfast.salm2map(D_alpha_modes.view(np.ndarray), 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1), spin_weight=0
    )

    # transform the Moreschi supermomentum
    PsiM_Grid_interp = (PsiM_Grid_interp - D_alpha_modes_Grid) / K_Grid_interp**3

    PsiM_interp = spinsfast.map2salm(PsiM_Grid_interp.view(np.ndarray), 0, ell_max)

    return PsiM_interp


def compute_alpha_perturbation(PsiM, M_Grid, K_Grid, ell_max):
    """From the Moreschi supermomentum transformation law,
    compute the supertranslation that maps to the superrest frame.

    This equation can be found in Eq (10) of
    https://doi.org/10.1063/1.532646.

    Parameters
    ----------
    PsiM: ModesTimeSeries
        Moreschi supermomentum computed from AsymptoticBondiData object.
    M_Grid: ndarray, complex, shape (..., Ntheta, Nphi)
        Bondi mass on the spherical grid.
    K_Grid: ndarray, complex, shape (..., Ntheta, Nphi)
        Conformal factor on the spherical grid.
    ell_max: int
        Maximum ell to use when converting data from SWSHs to data on the spherical grid.
    """
    PsiM_Grid = spinsfast.salm2map(PsiM.view(np.ndarray), 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1)
    PsiM_plus_M_K3_Grid = PsiM_Grid + M_Grid * K_Grid**3
    PsiM_plus_M_K3 = spinsfast.map2salm(PsiM_plus_M_K3_Grid.view(np.ndarray), 0, ell_max)

    alpha = 𝔇inverse(PsiM_plus_M_K3, ell_max)

    return spinsfast.salm2map(alpha, 0, ell_max, 2 * ell_max + 1, 2 * ell_max + 1).real


def supertranslation_to_map_to_superrest_frame(
    abd, target_PsiM=None, N_itr_max=10, rel_err_tol=1e-12, ell_max=12, print_conv=False
):
    """Determine the supertranslation needed to map an abd object to the superrest frame.

    This is found through an iterative solve; e.g., compute the supertranslation needed to minimize
    the Moreschi supermomentum according to Eq (10) of https://doi.org/10.1063/1.532646,
    transform the Moreschi supermomentum, and repeat until the supertranslation converges.

    Parameters
    ----------
    abd: AsymptoticBondiData
        AsymptoticBondiData object from which the Moreschi supermomentum will be computed.
    target_PsiM: ModesTimeSeries, defaults to None
        Target Moreschi supermomentum, e.g., the PN supermomentum which we may map the NR supermomentum to.
        Default is None, which equates to the target supermomentum being zero.
    N_itr_max: int, defaults to 10
        Maximum numebr of iterations to perform. Default is 10.
    rel_err_tol: float, defaults to 1e-12
        Minimum relativie error tolerance between transformation iterations. Default is 1e-12.
    ell_max: int, defaults to 12
        Maximum ell to use when converting data from SWSHs to data on the spherical grid. Default is 12.
    print_conv: bool, defaults to False
        Whether or not to print the termination criterion. Default is False.
    """
    alpha_Grid = np.zeros((2 * ell_max + 1, 2 * ell_max + 1), dtype=float)
    best_alpha_Grid = np.zeros((2 * ell_max + 1, 2 * ell_max + 1), dtype=float)

    PsiM = abd.supermomentum("Moreschi")

    itr = 0
    rel_err = np.inf
    rel_errs = [np.inf]
    while itr < N_itr_max and not rel_err < rel_err_tol:
        prev_alpha_Grid = sf.SWSH_grids.Grid(alpha_Grid.copy(), spin_weight=0)

        if itr == 0:
            PsiM_interp = compute_Moreschi_supermomentum(PsiM, alpha_Grid, ell_max)

            M_Grid, K_Grid = compute_bondi_rest_mass_and_conformal_factor(np.array(PsiM_interp), ell_max)

            if target_PsiM is not None:
                target_PsiM_Grid = WM_to_MT(target_PsiM).grid().real
                target_PsiM_Grid_interp = sf.SWSH_grids.Grid(
                    np.zeros((2 * target_PsiM.ell_max + 1, 2 * target_PsiM.ell_max + 1), dtype=float), spin_weight=0
                )
                for i in range(2 * target_PsiM.ell_max + 1):
                    for j in range(2 * target_PsiM.ell_max + 1):
                        alpha_i_j = prev_alpha_Grid[i, j]
                        target_PsiM_Grid_interp[i, j] = CubicSpline(target_PsiM.t, target_PsiM_Grid[:, i, j])(alpha_i_j)
                M_Grid = -target_PsiM_Grid_interp.view(np.ndarray)

        alpha_Grid += compute_alpha_perturbation(PsiM_interp, M_Grid, K_Grid, ell_max)

        PsiM_interp = compute_Moreschi_supermomentum(PsiM, alpha_Grid, ell_max)

        M_Grid, K_Grid = compute_bondi_rest_mass_and_conformal_factor(np.array(PsiM_interp), ell_max)

        if target_PsiM is not None:
            target_PsiM_Grid = WM_to_MT(target_PsiM).grid().real
            target_PsiM_Grid_interp = sf.SWSH_grids.Grid(
                np.zeros((2 * target_PsiM.ell_max + 1, 2 * target_PsiM.ell_max + 1), dtype=float), spin_weight=0
            )
            for i in range(2 * target_PsiM.ell_max + 1):
                for j in range(2 * target_PsiM.ell_max + 1):
                    alpha_i_j = prev_alpha_Grid[i, j]
                    target_PsiM_Grid_interp[i, j] = CubicSpline(target_PsiM.t, target_PsiM_Grid[:, i, j])(alpha_i_j)
            M_Grid = -target_PsiM_Grid_interp.view(np.ndarray)

            target_PsiM_interp = spinsfast.map2salm(target_PsiM_Grid_interp.view(np.ndarray), 0, ell_max)

            rel_err = np.linalg.norm(PsiM_interp[4:] - target_PsiM_interp[4:]) / np.linalg.norm(target_PsiM_interp[4:])
        else:
            rel_err = np.linalg.norm(PsiM_interp[4:])

        if rel_err < min(rel_errs):
            best_alpha_Grid = alpha_Grid.copy()
        rel_errs.append(rel_err)

        itr += 1

    if print_conv:
        if not itr < N_itr_max:
            print(
                f"supertranslation: maximum number of iterations reached; the min error was {min(np.array(rel_errs).flatten())}."
            )
        else:
            print(f"supertranslation: tolerance achieved in {itr} iterations!")

    supertranslation = spinsfast.map2salm(best_alpha_Grid.view(np.ndarray), 0, ell_max)
    supertranslation[0:4] = 0

    return scri.bms_transformations.BMSTransformation(supertranslation=supertranslation), rel_errs


def transformation_from_CoM_charge(G, t):
    """Obtain the space translation and boost velocity from the center-of-mass charge.

    This is defined in Eq (18) of https://journals.aps.org/prd/abstract/10.1103/PhysRevD.104.024051.

    Parameters
    ----------
    G: ndarray, real, shape (..., 3)
        Center-of-mass charge.
    t: ndarray, real
        Time array corresponding to the size of the center-of-mass charge.
    """
    polynomial_fit = np.polyfit(t, G, deg=1)

    CoM_transformation = scri.bms_transformations.BMSTransformation(
        supertranslation=-np.insert(sf.vector_as_ell_1_modes(polynomial_fit[1]), 0, 0),
        boost_velocity=polynomial_fit[0],
        order=["supertranslation", "boost_velocity", "frame_rotation"],
    )

    return CoM_transformation


def com_transformation_to_map_to_superrest_frame(abd, N_itr_max=10, rel_err_tol=1e-12, print_conv=False):
    """Determine the space translation and boost needed to map an abd object to the superrest frame.

    These are found through an iterative solve; e.g., compute the transformations needed to minimize
    the center-of-mass charge, transform the abd object, and repeat until the transformations converge.

    This function can also be used to find just the space translation or the boost velocity, rather than both.

    Parameters
    ----------
    abd: AsymptoticBondiData
        AsymptoticBondiData object from which the Moreschi supermomentum will be computed.
    N_itr_max: int, defaults to 10
        Maximum number of iterations to perform. Default is 10.
    rel_err_tol: float, defaults to 1e-12
        Minimum relativie error tolerance between transformation iterations. Default is 1e-12.
    print_conv: bool, defaults to False
        Whether or not to print the termination criterion. Default is False.
    """
    CoM_transformation = scri.bms_transformations.BMSTransformation()
    best_CoM_transformation = scri.bms_transformations.BMSTransformation()

    itr = 0
    rel_err = np.inf
    rel_errs = [np.inf]
    while itr < N_itr_max and not rel_err < rel_err_tol:
        if itr == 0:
            abd_prime = abd.copy()
            G_prime = abd_prime.bondi_CoM_charge() / abd_prime.bondi_four_momentum()[:, 0, None]

        new_CoM_transformation = transformation_from_CoM_charge(G_prime, abd_prime.t)
        CoM_transformation = (new_CoM_transformation * CoM_transformation).reorder(
            ["supertranslation", "frame_rotation", "boost_velocity"]
        )
        # remove proper supertranslation and frame rotation components
        CoM_transformation.supertranslation[4:] *= 0
        CoM_transformation.frame_rotation = quaternion.quaternion(1, 0, 0, 0)

        abd_prime = abd.transform(
            supertranslation=CoM_transformation.supertranslation,
            frame_rotation=CoM_transformation.frame_rotation.components,
            boost_velocity=CoM_transformation.boost_velocity,
        )

        G_prime = abd_prime.bondi_CoM_charge() / abd_prime.bondi_four_momentum()[:, 0, None]

        rel_err = integrate(np.linalg.norm(G_prime, axis=-1), abd_prime.t)[-1] / (abd_prime.t[-1] - abd_prime.t[0])
        if rel_err < min(rel_errs):
            best_CoM_transformation = CoM_transformation.copy()
        rel_errs.append(rel_err)

        itr += 1

    if print_conv:
        if not itr < N_itr_max:
            print(f"CoM: maximum number of iterations reached; the min error was {min(rel_errs)}.")
        else:
            print(f"CoM: tolerance achieved in {itr} iterations!")

    return best_CoM_transformation, rel_errs


def rotation_from_spin_charge(chi, t):
    """Obtain the rotation from the remnant BH's spin vector.
    This finds the rotation that aligns the z-component of the spin vector with the z-axis.
    Parameters
    ----------
    chi: ndarray, real, shape (..., 3)
        Remnant BH's spin vector.
    t: ndarray, real
        Time array corresponding to the size of the spin vector.
    """
    chi_f = quaternion.quaternion(*chi[np.argmin(abs(t))]).normalized()
    q = (1 - chi_f * quaternion.z).normalized().components
    return scri.bms_transformations.BMSTransformation(frame_rotation=q)


def rotation_from_vectors(vector, target_vector, t=None):
    """Obtain the rotation from the two vectors.

    This finds the rotation that best aligns two vectors over the time interval t.

    Parameters
    ----------
    vector: ndarray, real, shape (..., 3)
        Angular vector.
    target_vector: ndarray, real, shape (..., 3)
        Target vector.
    t: ndarray, real
        Time array corresponding to the size of the spin vector. Default is None.
    """
    q = quaternion.optimal_alignment_in_Euclidean_metric(vector, target_vector, t=t).inverse()

    return scri.bms_transformations.BMSTransformation(frame_rotation=q.components)


def rotation_to_map_to_superrest_frame(abd, target_strain=None, N_itr_max=10, rel_err_tol=1e-12, print_conv=False):
    """Determine the rotation needed to map an abd object to the superrest frame.

    This is found through an iterative solve; e.g., compute the transformation needed to align
    the spin vector charge with the z-axis or align the angular velocity with a target angular velocity,
    transform the abd object, and repeat until the transformation converges.

    Note that the angular momentum charge is aligned with either the
    positive or negative z-axis, depending on which it is initially closest to.

    If target_h is not None, then instead find the rotation that aligns the
    angular velocity vector of the abd object to the angular velocity vector
    of the target_h input.

    Parameters
    ----------
    abd: AsymptoticBondiData
        AsymptoticBondiData object from which the Moreschi supermomentum will be computed.
    target_strain: WaveformModes, optional
        WaveformModes object from which the target angular velocity will be computed. Default is None.
    N_itr_max: int, defaults to 10
        Maximum number of iterations to perform. Default is 10.
    rel_err_tol: array_like, defaults to 1e-12
        First value is minimum relative error tolerance between transformation iterations; second value is the
        minimum relative error tolerance between the NR angular velocity and the target angular velocity.
        Default is 1e-12.
    print_conv: bool, defaults to False
        Whether or not to print the termination criterion. Default is False.
    """
    rotation_transformation = scri.bms_transformations.BMSTransformation()
    best_rotation_transformation = scri.bms_transformations.BMSTransformation()

    # if there is no target_strain, just map the spin charge to the z-axis;
    # otherwise, align the angular momentum fluxes over the full window.
    if target_strain is not None:
        target_news = MT_to_WM(
            scri.ModesTimeSeries(
                sf.SWSH_modes.Modes(
                    target_strain.data_dot,
                    spin_weight=-2,
                    ell_min=2,
                    ell_max=target_strain.ell_max,
                    multiplication_truncator=max,
                ),
                time=target_strain.t,
            ).dot,
            dataType=scri.hdot,
        )

        target_omega = target_news.angular_velocity()
        target_omega_spline = CubicSpline(
            target_strain.t, target_omega / np.linalg.norm(target_omega, axis=-1)[:, None]
        )

        itr = 0
        rel_err = np.inf
        rel_errs = [np.inf]
        while itr < N_itr_max and not rel_err < rel_err_tol:
            if itr == 0:
                abd_prime = abd.copy()
                news = MT_to_WM(2.0 * abd.sigma.bar.dot, dataType=scri.hdot)
                omega = news.angular_velocity()
                omega = omega / np.linalg.norm(omega, axis=-1)[:, None]

            rotation_transformation = (
                rotation_from_vectors(omega, target_omega_spline(news.t), news.t) * rotation_transformation
            ).reorder(["supertranslation", "frame_rotation", "boost_velocity"])
            # remove supertranslation and CoM components
            rotation_transformation.supertranslation *= 0
            rotation_transformation.boost_velocity *= 0

            abd_prime = abd.transform(frame_rotation=rotation_transformation.frame_rotation.components)

            news = MT_to_WM(2.0 * abd_prime.sigma.bar.dot, dataType=scri.hdot)
            omega = news.angular_velocity()
            omega = omega / np.linalg.norm(omega, axis=-1)[:, None]

            rel_err = integrate(np.linalg.norm(omega - target_omega_spline(news.t), axis=-1), news.t)[-1] / (
                abd_prime.t[-1] - abd_prime.t[0]
            )
            if rel_err < min(rel_errs):
                best_rotation_transformation = rotation_transformation.copy()
            rel_errs.append(rel_err)

            itr += 1
    else:
        itr = 0
        rel_err = np.inf
        rel_errs = [np.inf]
        while itr < N_itr_max and not rel_err < rel_err_tol:
            if itr == 0:
                abd_prime = abd.copy()
                chi_prime = abd_prime.bondi_dimensionless_spin()
                chi_prime = chi_prime / np.linalg.norm(chi_prime, axis=-1)[:, None]

            rotation_transformation = (
                rotation_from_spin_charge(chi_prime, abd_prime.t) * rotation_transformation
            ).reorder(["supertranslation", "frame_rotation", "boost_velocity"])
            # remove supertranslation and CoM components
            rotation_transformation.supertranslation *= 0
            rotation_transformation.boost_velocity *= 0

            abd_prime = abd.transform(frame_rotation=rotation_transformation.frame_rotation.components)

            chi_prime = abd_prime.bondi_dimensionless_spin()
            chi_prime = chi_prime / np.linalg.norm(chi_prime, axis=-1)[:, None]

            rel_err = integrate(np.linalg.norm(chi_prime - [[0, 0, 1]] * abd_prime.t.size, axis=-1), abd_prime.t)[
                -1
            ] / (abd_prime.t[-1] - abd_prime.t[0])
            if rel_err < min(rel_errs):
                best_rotation_transformation = rotation_transformation.copy()
            rel_errs.append(rel_err)

            itr += 1

    if print_conv:
        if not itr < N_itr_max:
            print(
                f"rotation: maximum number of iterations reached; the min error was {min(np.array(rel_errs).flatten())}."
            )
        else:
            print(f"rotation: tolerance achieved in {itr} iterations!")

    return best_rotation_transformation, rel_errs


def time_translation(abd, t_0=0):
    """Time translate an abd object.

    This is necessary because creating a copy
    of an abd object and then changing it's time variable does not change
    the time variable of the waveform variables.
    """
    abd_prime = scri.asymptotic_bondi_data.AsymptoticBondiData(abd.t, abd.ell_max)

    abd_prime.sigma = abd.sigma
    abd_prime.psi4 = abd.psi4
    abd_prime.psi3 = abd.psi3
    abd_prime.psi2 = abd.psi2
    abd_prime.psi1 = abd.psi1
    abd_prime.psi0 = abd.psi0

    abd_prime.t -= t_0

    return abd_prime


def rotation(abd, phi=0):
    """Rotate an abd object.

    This is faster than using abd.transform().
    """
    q = quaternion.from_rotation_vector(-phi * np.array([0, 0, 1]))

    h = MT_to_WM(2.0 * abd.sigma.bar, False, scri.h)
    Psi4 = MT_to_WM(0.5 * (-np.sqrt(2)) ** 4 * abd.psi4, False, scri.psi4)
    Psi3 = MT_to_WM(0.5 * (-np.sqrt(2)) ** 3 * abd.psi3, False, scri.psi3)
    Psi2 = MT_to_WM(0.5 * (-np.sqrt(2)) ** 2 * abd.psi2, False, scri.psi2)
    Psi1 = MT_to_WM(0.5 * (-np.sqrt(2)) ** 1 * abd.psi1, False, scri.psi1)
    Psi0 = MT_to_WM(0.5 * (-np.sqrt(2)) ** 0 * abd.psi0, False, scri.psi0)

    h.rotate_physical_system(q)
    Psi4.rotate_physical_system(q)
    Psi3.rotate_physical_system(q)
    Psi2.rotate_physical_system(q)
    Psi1.rotate_physical_system(q)
    Psi0.rotate_physical_system(q)

    abd_rot = abd.copy()
    abd_rot.sigma = 0.5 * WM_to_MT(h).bar
    abd_rot.psi4 = 2 * (-1.0 / np.sqrt(2)) ** 4 * WM_to_MT(Psi4)
    abd_rot.psi3 = 2 * (-1.0 / np.sqrt(2)) ** 3 * WM_to_MT(Psi3)
    abd_rot.psi2 = 2 * (-1.0 / np.sqrt(2)) ** 2 * WM_to_MT(Psi2)
    abd_rot.psi1 = 2 * (-1.0 / np.sqrt(2)) ** 1 * WM_to_MT(Psi1)
    abd_rot.psi0 = 2 * (-1.0 / np.sqrt(2)) ** 0 * WM_to_MT(Psi0)

    return abd_rot


def rel_err_for_abd_in_superrest(abd, target_PsiM, target_strain):
    G = abd.bondi_CoM_charge() / abd.bondi_four_momentum()[:, 0, None]
    rel_err_CoM_transformation = integrate(np.linalg.norm(G, axis=-1), abd.t)[-1] / (abd.t[-1] - abd.t[0])

    if target_strain is not None:
        target_news = MT_to_WM(
            scri.ModesTimeSeries(
                sf.SWSH_modes.Modes(
                    target_strain.data_dot,
                    spin_weight=-2,
                    ell_min=2,
                    ell_max=target_strain.ell_max,
                    multiplication_truncator=max,
                ),
                time=target_strain.t,
            ).dot,
            dataType=scri.hdot,
        )

        target_omega = target_news.angular_velocity()
        target_omega_spline = CubicSpline(
            target_strain.t, target_omega / np.linalg.norm(target_omega, axis=-1)[:, None]
        )

        news = MT_to_WM(2.0 * abd.sigma.bar.dot, dataType=scri.hdot)
        omega = news.angular_velocity()
        omega = omega / np.linalg.norm(omega, axis=-1)[:, None]

        rel_err_rotation = integrate(np.linalg.norm(omega - target_omega_spline(news.t), axis=-1), news.t)[-1] / (
            abd.t[-1] - abd.t[0]
        )
    else:
        spin = abd.bondi_dimensionless_spin()
        spin = spin / np.linalg.norm(spin, axis=-1)[:, None]
        rel_err_rotation = integrate(np.linalg.norm(spin - [[0, 0, 1]] * abd.t.size, axis=-1), abd.t)[-1] / (
            abd.t[-1] - abd.t[0]
        )

    if target_PsiM is not None:
        rel_err_PsiM = np.linalg.norm(
            np.array(abd.supermomentum("Moreschi"))[np.argmin(abs(abd.t - 0)), 4:]
            - np.array(target_PsiM.data)[np.argmin(abs(target_PsiM.t - 0)), 4:]
        )
    else:
        rel_err_PsiM = np.linalg.norm(np.array(abd.supermomentum("Moreschi"))[np.argmin(abs(abd.t - 0)), 4:])

    return rel_err_CoM_transformation, rel_err_rotation, rel_err_PsiM


def map_to_superrest_frame(
    self,
    t_0=0,
    target_PsiM_input=None,
    target_strain_input=None,
    padding_time=250,
    N_itr_maxes={
        "superrest": 2,
        "CoM_transformation": 10,
        "rotation": 10,
        "supertranslation": 10,
    },
    rel_err_tols={"CoM_transformation": 1e-12, "rotation": 1e-12, "supertranslation": 1e-12},
    order=["supertranslation", "rotation", "CoM_transformation"],
    ell_max=None,
    alpha_ell_max=None,
    fix_time_phase_freedom=False,
    modes=None,
    print_conv=False,
):
    """Transform an abd object to the superrest frame.

    This computes the transformations necessary to map an abd object to the superrest frame
    by iteratively minimizing various BMS charges at a certain time. This can be performed either
    without a target (in which case it maps to the superrest frame at the input time t_0) or
    with a target (in which case it maps to the frame of the target). For example, one could
    supply a t_0 which is toward the end of the simulation in which case this code would map to
    the superrest frame of the remnant BH. Or, one could supply a target PN strain and supermomentum and
    a t_0 during the early inspiral in which case this code would map to the PN BMS frame.

    This wholly fixes the BMS frame of the abd object up to a
    time translation and a phase rotation.

    Parameters
    ----------
    t_0 : float, optional
        When to map to the superrest frame.
        Default is 0.
    target_PsiM_input : WaveformModes, optional
        Target Moreschi supermomentum to map to.
        Default is 0.
    target_strain_input : WaveformModes, optional
        Target strain used to constrain the rotation
        freedom via the angular momentum flux.
        Default is aligning to the z-axis.
    padding_time : float, optional
        Amount by which to pad around t_0 to speed up computations, i.e.,
        distance from t_0 in each direction to be included in self.interpolate(...).
        This also determines the range over which certain BMS charges will be computed.
        Default is 250.
    N_itr_maxes : dict, optional
        Maximum number of iterations to perform for each transformation.
        For 'superrest', this is the number of iterations to use for the superrest procedure.
        For the other options, these are the number of iterations to use for finding each individual transformation.
        Default is
        N_itr_maxes = {
            'superrest':          2,
            'CoM_transformation': 10,
            'rotation':           10,
            'supertranslation':   10,
        }.
    rel_err_tols : dict, optional
        Relative error tolerances for each transformation.
        Default is
        rel_err_tols = {
            'CoM_transformation': 1e-12,
            'rotation':           1e-12,
            'supertranslation':   1e-12
        }.
    order : list, optional
        Order in which to solve for the BMS transformations.
        Default is ["rotation", "CoM_transformation", "supertranslation"].
    ell_max : int, optional
        Maximum ell to use for SWSH/Grid transformations.
        Default is self.ell_max.
    alpha_ell_max : int, optional
        Maximum ell of the supertranslation to use.
        Default is self.ell_max.
    fix_time_phase_freedom : bool, optional
        Whether or not to fix the time and phase freedom using a 2d minimization scheme.
        Default is True.
    modes : list, optional
        List of modes to include when performing the 2d alignment.
        Default is every mode.
    print_conv: bool, defaults to False
        Whether or not to print the termination criterion. Default is False.

    Returns
    -------
    abd_prime : AsymptoticBondiData
        Result of self.transform(...) where the input transformations are
        the transformations found in the BMSTransformations object.
    transformations : BMSTransformation
        BMS transformation to map to the target BMS frame.

    """
    abd = self.copy()

    if target_strain_input is not None:
        target_strain = target_strain_input.copy()
        target_strain.t -= t_0
    else:
        target_strain = None

    if target_PsiM_input is not None:
        target_PsiM = target_PsiM_input.copy()
        target_PsiM.t -= t_0
    else:
        target_PsiM = None

    if ell_max is None:
        ell_max = abd.ell_max

    if alpha_ell_max is None:
        alpha_ell_max = ell_max

    abd_interp = abd.interpolate(
        abd.t[np.argmin(abs(abd.t - (t_0 - (padding_time + 200)))) : np.argmin(abs(abd.t - (t_0 + (padding_time + 200)))) + 1]
    )

    # apply a time translation so that we're mapping
    # to the superrest frame at u = 0
    time_translation = scri.bms_transformations.BMSTransformation(supertranslation=[sf.constant_as_ell_0_mode(t_0)])
    BMS_transformation = time_translation * scri.bms_transformations.BMSTransformation().reorder(
        ["supertranslation", "frame_rotation", "boost_velocity"]
    )

    itr = 0
    rel_err = [np.inf, np.inf, np.inf]
    rel_errs = [[np.inf, np.inf, np.inf]]
    best_rel_err = [np.inf, np.inf, np.inf]
    while itr < N_itr_maxes["superrest"]:
        if type(rel_err) == tuple:
            if (
                    rel_err[0] < rel_err_tols["CoM_transformation"]
                    and rel_err[1] < rel_err_tols["rotation"]
                    and rel_err[2] < rel_err_tols["supertranslation"]
            ):
                break
        else:
            pass
            
        if itr == 0:
            abd_interp_prime = abd_interp.transform(
                supertranslation=BMS_transformation.supertranslation,
                frame_rotation=BMS_transformation.frame_rotation.components,
                boost_velocity=BMS_transformation.boost_velocity,
            )

        for transformation in order:
            if transformation == "supertranslation":
                new_transformation, supertranslation_rel_errs = supertranslation_to_map_to_superrest_frame(
                    abd_interp_prime,
                    target_PsiM,
                    N_itr_max=N_itr_maxes["supertranslation"],
                    rel_err_tol=rel_err_tols["supertranslation"],
                    ell_max=ell_max,
                    print_conv=print_conv,
                )
            elif transformation == "rotation":
                new_transformation, rot_rel_errs = rotation_to_map_to_superrest_frame(
                    abd_interp_prime,
                    target_strain=target_strain,
                    N_itr_max=N_itr_maxes["rotation"],
                    rel_err_tol=rel_err_tols["rotation"],
                    print_conv=print_conv,
                )
            elif transformation == "CoM_transformation":
                new_transformation, CoM_rel_errs = com_transformation_to_map_to_superrest_frame(
                    abd_interp_prime,
                    N_itr_max=N_itr_maxes["CoM_transformation"],
                    rel_err_tol=rel_err_tols["CoM_transformation"],
                    print_conv=print_conv,
                )
            elif transformation == "time_phase":
                if target_strain is not None:
                    strain_interp_prime = scri.asymptotic_bondi_data.map_to_superrest_frame.MT_to_WM(
                        2.0 * abd_interp_prime.sigma.bar, dataType=scri.h
                    )

                    rel_err, _, res = align2d(
                        MT_to_WM(WM_to_MT(strain_interp_prime), sxs_version=True),
                        MT_to_WM(WM_to_MT(target_strain), sxs_version=True),
                        0 - padding_time,
                        0 + padding_time,
                        n_brute_force_δt=None,
                        n_brute_force_δϕ=None,
                        include_modes=modes,
                        nprocs=4,
                    )

                    new_transformation = scri.bms_transformations.BMSTransformation(
                        supertranslation=[sf.constant_as_ell_0_mode(res.x[0])],
                        frame_rotation=quaternion.from_rotation_vector(res.x[1] * np.array([0, 0, 1])).components,
                    )

            BMS_transformation = (new_transformation * BMS_transformation).reorder(
                ["supertranslation", "frame_rotation", "boost_velocity"]
            )

            abd_interp_prime = abd_interp.transform(
                supertranslation=BMS_transformation.supertranslation,
                frame_rotation=BMS_transformation.frame_rotation.components,
                boost_velocity=BMS_transformation.boost_velocity,
            )

        if target_strain is not None and order[-1] == "time_phase":
            # rel_err is obtained from align2d, so do nothing
            pass
        else:
            rel_err = rel_err_for_abd_in_superrest(abd_interp_prime, target_PsiM, target_strain)
            
        if np.mean(rel_err) < min([np.mean(r) for r in rel_errs]):
            best_BMS_transformation = BMS_transformation.copy()
            best_rel_err = rel_err
        rel_errs.append(rel_err)

        itr += 1

    if print_conv:
        if not itr < N_itr_maxes["superrest"]:
            print(f"superrest: maximum number of iterations reached; the min error was {best_rel_err}.")
        else:
            print(f"superrest: tolerance achieved in {itr} iterations!")

    # undo the time translation
    best_BMS_transformation = (time_translation.inverse() * best_BMS_transformation).reorder(
        ["supertranslation", "frame_rotation", "boost_velocity"]
    )

    # transform abd
    abd_prime = abd.transform(
        supertranslation=best_BMS_transformation.supertranslation,
        frame_rotation=best_BMS_transformation.frame_rotation.components,
        boost_velocity=best_BMS_transformation.boost_velocity,
    )

    return abd_prime, best_BMS_transformation, best_rel_err



import pytest
import numpy as np

import scri
import spherical_functions as sf
from scri import bms_transformations

from conftest import kerr_schild

ABD = scri.AsymptoticBondiData


def rotation_matrix(q):
    s, x, y, z = q.components
    R_matrix = [
        [1, 0, 0, 0],
        [0, 1 - 2 * y**2 - 2 * z**2, 2 * x * y - 2 * s * z, 2 * x * z + 2 * s * y],
        [0, 2 * x * y + 2 * s * z, 1 - 2 * x**2 - 2 * z**2, 2 * y * z - 2 * s * x],
        [0, 2 * x * z - 2 * s * y, 2 * y * z + 2 * s * x, 1 - 2 * x**2 - 2 * y**2],
    ]

    return R_matrix


def boost_matrix(v):
    v_norm = np.linalg.norm(v)
    g = 1 / np.sqrt(1 - v_norm**2)
    B_matrix = [
        [g, -g * v[0], -g * v[1], -g * v[2]],
        [
            -g * v[0],
            1 + (g - 1) * v[0] ** 2 / v_norm**2,
            (g - 1) * v[0] * v[1] / v_norm**2,
            (g - 1) * v[0] * v[2] / v_norm**2,
        ],
        [
            -g * v[1],
            (g - 1) * v[1] * v[0] / v_norm**2,
            1 + (g - 1) * v[1] ** 2 / v_norm**2,
            (g - 1) * v[1] * v[2] / v_norm**2,
        ],
        [
            -g * v[2],
            (g - 1) * v[2] * v[0] / v_norm**2,
            (g - 1) * v[2] * v[1] / v_norm**2,
            1 + (g - 1) * v[2] ** 2 / v_norm**2,
        ],
    ]

    return B_matrix


def test_Lorentz_reorder_consistency():
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-2
    L = bms_transformations.LorentzTransformation(
        frame_rotation=q.components, boost_velocity=v, order=["frame_rotation", "boost_velocity"]
    )

    L_reordered = L.reorder(output_order=L.order[::-1])

    assert L.is_close_to(L_reordered.reorder(output_order=L.order))


def test_Lorentz_reorder():
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-2
    L = bms_transformations.LorentzTransformation(
        frame_rotation=q.components, boost_velocity=v, order=["frame_rotation", "boost_velocity"]
    )

    L_reordered = L.reorder(output_order=L.order[::-1])

    vq = np.linalg.multi_dot([boost_matrix(v), rotation_matrix(q)])

    L_reordered_matrix = np.linalg.multi_dot(
        [rotation_matrix(L_reordered.frame_rotation), boost_matrix(L_reordered.boost_velocity)]
    )

    assert np.allclose(vq, L_reordered_matrix)


def test_frame_rotation_inverse():
    q = np.quaternion(1, 2, 3, 4).normalized()
    L = bms_transformations.LorentzTransformation(frame_rotation=q.components)

    assert np.allclose(q.inverse().components, L.inverse().frame_rotation.components)


def test_boost_velocity_inverse():
    v = np.array([1, 2, 3]) * 1e-2
    L = bms_transformations.LorentzTransformation(boost_velocity=v)

    assert np.allclose(-v, L.inverse().boost_velocity)


def test_Lorentz_inverse_composition_consistency():
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-2
    L = bms_transformations.LorentzTransformation(
        frame_rotation=q.components, boost_velocity=v, order=["frame_rotation", "boost_velocity"]
    )
    L_inv = L.inverse()

    assert (L * L_inv).is_close_to(bms_transformations.LorentzTransformation())
    assert (L_inv * L).is_close_to(bms_transformations.LorentzTransformation())


def test_Lorentz_inverse():
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-2
    L = bms_transformations.LorentzTransformation(
        frame_rotation=q.components, boost_velocity=v, order=["frame_rotation", "boost_velocity"]
    )
    L_inv = bms_transformations.LorentzTransformation(
        frame_rotation=q.inverse().components, boost_velocity=-v, order=["boost_velocity", "frame_rotation"]
    )

    assert L_inv.is_close_to(L.inverse(output_order=["boost_velocity", "frame_rotation"]))


def test_frame_rotation_composition():
    q1 = np.quaternion(1, 2, 3, 4).normalized()
    q2 = np.quaternion(5, -6, 7, -8).normalized()
    L1 = bms_transformations.LorentzTransformation(frame_rotation=q1.components)
    L2 = bms_transformations.LorentzTransformation(frame_rotation=q2.components)

    q1q2 = q2 * q1
    q2q1 = q1 * q2

    L1L2 = L1 * L2
    L2L1 = L2 * L1

    assert np.allclose(q1q2.components, L1L2.frame_rotation.components)
    assert np.allclose(q2q1.components, L2L1.frame_rotation.components)


def test_boost_velocity_composition():
    v1 = np.array([1, 2, 3]) * 1e-2
    v2 = np.array([-4, 5, -6]) * 1e-2

    L1 = bms_transformations.LorentzTransformation(boost_velocity=v1)
    L2 = bms_transformations.LorentzTransformation(boost_velocity=v2)

    # the following Wigner rotation calculation
    # is from Eqs. (65) - (70) of arXiv:1102.2001
    g1 = 1 / np.sqrt(1 - np.linalg.norm(v1) ** 2)
    g2 = 1 / np.sqrt(1 - np.linalg.norm(v2) ** 2)
    g12 = g1 * g2 * (1 + np.dot(v1, v2))
    v1v2 = (v1 + g2 * v2 + (g2 - 1) * np.dot(v1, v2) * v2 / np.linalg.norm(v2) ** 2) / (g2 * (1 + np.dot(v1, v2)))
    v2v1 = (v2 + g1 * v1 + (g1 - 1) * np.dot(v2, v1) * v1 / np.linalg.norm(v1) ** 2) / (g1 * (1 + np.dot(v2, v1)))

    theta = np.arccos((1 + g1 + g2 + g12) ** 2 / ((1 + g1) * (1 + g2) * (1 + g12)) - 1)
    v1v2_q = np.quaternion(
        np.cos(theta / 2), *(np.cross(v1, v2) / np.linalg.norm(np.cross(v1, v2)) * np.sin(theta / 2))
    )
    v2v1_q = np.quaternion(
        np.cos(theta / 2), *(np.cross(v2, v1) / np.linalg.norm(np.cross(v2, v1)) * np.sin(theta / 2))
    )

    L1L2 = L1 * L2
    L2L1 = L2 * L1

    assert np.allclose(v1v2, L1L2.boost_velocity)
    assert np.allclose(v1v2_q.components, L1L2.frame_rotation.components)
    assert np.allclose(v2v1, L2L1.boost_velocity)
    assert np.allclose(v2v1_q.components, L2L1.frame_rotation.components)


def test_Lorentz_composition():
    q1 = np.quaternion(1, 2, 3, 4).normalized()
    v1 = np.array([1, 2, 3]) * 1e-2
    q2 = np.quaternion(5, -6, 7, -8).normalized()
    v2 = np.array([-4, 5, -6]) * 1e-2

    L1 = bms_transformations.LorentzTransformation(frame_rotation=q1.components, boost_velocity=v1)
    L2 = bms_transformations.LorentzTransformation(frame_rotation=q2.components, boost_velocity=v2)

    q1v1q2v2 = np.linalg.multi_dot([boost_matrix(v2), rotation_matrix(q2), boost_matrix(v1), rotation_matrix(q1)])
    q2v2q1v1 = np.linalg.multi_dot([boost_matrix(v1), rotation_matrix(q1), boost_matrix(v2), rotation_matrix(q2)])

    L1L2 = L1 * L2
    L1L2_matrix = np.linalg.multi_dot([boost_matrix(L1L2.boost_velocity), rotation_matrix(L1L2.frame_rotation)])

    L2L1 = L2 * L1
    L2L1_matrix = np.linalg.multi_dot([boost_matrix(L2L1.boost_velocity), rotation_matrix(L2L1.frame_rotation)])

    assert np.allclose(q1v1q2v2, L1L2_matrix)
    assert np.allclose(q2v2q1v1, L2L1_matrix)


def test_Lorentz_abd_inverse():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=500)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-4

    Lorentz = bms_transformations.LorentzTransformation(
        frame_rotation=q.components, boost_velocity=v, order=["frame_rotation", "boost_velocity"]
    )

    Lorentz_inv = Lorentz.inverse(output_order=["frame_rotation", "boost_velocity"])

    abd_prime = abd.transform(frame_rotation=Lorentz.frame_rotation.components, boost_velocity=Lorentz.boost_velocity)

    abd_check = abd_prime.transform(
        frame_rotation=Lorentz_inv.frame_rotation.components, boost_velocity=Lorentz_inv.boost_velocity
    )

    abd_interp = abd.interpolate(abd_check.t)

    assert np.allclose(
        np.array(
            [abd_interp.sigma, abd_interp.psi4, abd_interp.psi3, abd_interp.psi2, abd_interp.psi1, abd_interp.psi0]
        ),
        np.array([abd_check.sigma, abd_check.psi4, abd_check.psi3, abd_check.psi2, abd_check.psi1, abd_check.psi0]),
    )


def test_Lorentz_abd_composition():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=500)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    q1 = np.quaternion(1, 2, 3, 4).normalized()
    v1 = np.array([1, 2, 3]) * 1e-4
    q2 = np.quaternion(5, -6, 7, -8).normalized()
    v2 = np.array([-4, 5, -6]) * 1e-4

    Lorentz1 = bms_transformations.LorentzTransformation(
        frame_rotation=q1.components, boost_velocity=v1, order=["frame_rotation", "boost_velocity"]
    )

    Lorentz2 = bms_transformations.LorentzTransformation(
        frame_rotation=q2.components, boost_velocity=v2, order=["frame_rotation", "boost_velocity"]
    )

    Lorentz_composed = Lorentz2 * Lorentz1

    abd1 = abd.transform(frame_rotation=Lorentz1.frame_rotation.components, boost_velocity=Lorentz1.boost_velocity)

    abd2 = abd1.transform(frame_rotation=Lorentz2.frame_rotation.components, boost_velocity=Lorentz2.boost_velocity)

    abd_composed = abd.transform(
        frame_rotation=Lorentz_composed.frame_rotation.components, boost_velocity=Lorentz_composed.boost_velocity
    )

    abd2_interp = abd2.interpolate(
        abd.t[
            np.argmin(abs(abd.t - max(abd2.t[0], abd_composed.t[0]))) : np.argmin(
                abs(abd.t - min(abd2.t[-1], abd_composed.t[-1]))
            )
            + 1
        ]
    )
    abd_composed_interp = abd_composed.interpolate(
        abd.t[
            np.argmin(abs(abd.t - max(abd2.t[0], abd_composed.t[0]))) : np.argmin(
                abs(abd.t - min(abd2.t[-1], abd_composed.t[-1]))
            )
            + 1
        ]
    )

    assert np.allclose(
        np.array(
            [
                abd2_interp.sigma,
                abd2_interp.psi4,
                abd2_interp.psi3,
                abd2_interp.psi2,
                abd2_interp.psi1,
                abd2_interp.psi0,
            ]
        ),
        np.array(
            [
                abd_composed_interp.sigma,
                abd_composed_interp.psi4,
                abd_composed_interp.psi3,
                abd_composed_interp.psi2,
                abd_composed_interp.psi1,
                abd_composed_interp.psi0,
            ]
        ),
    )


def test_BMS_reorder_consistency():
    S = np.array([1, 2 + 4j, 3, -2 + 4j, 7 - 5j, -3 - 2j, 4, 3 - 2j, 7 + 5j]) * 1e-3
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-4

    BMS = bms_transformations.BMSTransformation(supertranslation=S, frame_rotation=q.components, boost_velocity=v)

    check1 = BMS.is_close_to(BMS.reorder(["supertranslation", "frame_rotation", "boost_velocity"]).reorder(BMS.order))
    check2 = BMS.is_close_to(BMS.reorder(["frame_rotation", "supertranslation", "boost_velocity"]).reorder(BMS.order))
    check3 = BMS.is_close_to(BMS.reorder(["frame_rotation", "boost_velocity", "supertranslation"]).reorder(BMS.order))
    check4 = BMS.is_close_to(BMS.reorder(["supertranslation", "boost_velocity", "frame_rotation"]).reorder(BMS.order))
    check5 = BMS.is_close_to(BMS.reorder(["boost_velocity", "supertranslation", "frame_rotation"]).reorder(BMS.order))
    check6 = BMS.is_close_to(BMS.reorder(["boost_velocity", "frame_rotation", "supertranslation"]).reorder(BMS.order))

    check_final = BMS.is_close_to(
        BMS.reorder(["frame_rotation", "supertranslation", "boost_velocity"])
        .reorder(["supertranslation", "boost_velocity", "frame_rotation"])
        .reorder(["boost_velocity", "supertranslation", "frame_rotation"])
        .reorder(["frame_rotation", "boost_velocity", "supertranslation"])
        .reorder(["boost_velocity", "frame_rotation", "supertranslation"])
        .reorder(BMS.order)
    )

    assert check1
    assert check2
    assert check3
    assert check4
    assert check5
    assert check6
    assert check_final


def test_BMS_inverse_composition_consistency():
    S = np.array([1, 2 + 4j, 3, -2 + 4j, 7 - 5j, -3 - 2j, 4, 3 - 2j, 7 + 5j]) * 1e-3
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-4

    BMS = bms_transformations.BMSTransformation(supertranslation=S, frame_rotation=q.components, boost_velocity=v)
    BMS_inv = BMS.inverse()

    assert (BMS * BMS_inv).is_close_to(bms_transformations.BMSTransformation())
    assert (BMS_inv * BMS).is_close_to(bms_transformations.BMSTransformation())


def test_BMS_abd_inverse():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=500)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    S = np.array([1, 2 + 4j, 3, -2 + 4j, 7 - 5j, -3 - 2j, 4, 3 - 2j, 7 + 5j]) * 1e-3
    q = np.quaternion(1, 2, 3, 4).normalized()
    v = np.array([1, 2, 3]) * 1e-4

    BMS = bms_transformations.BMSTransformation(
        supertranslation=S,
        frame_rotation=q.components,
        boost_velocity=v,
        order=["supertranslation", "frame_rotation", "boost_velocity"],
    )

    BMS_inv = BMS.inverse(output_order=["supertranslation", "frame_rotation", "boost_velocity"])

    abd_prime = abd.transform(
        supertranslation=BMS.supertranslation,
        frame_rotation=BMS.frame_rotation.components,
        boost_velocity=BMS.boost_velocity,
    )

    abd_check = abd_prime.transform(
        supertranslation=BMS_inv.supertranslation,
        frame_rotation=BMS_inv.frame_rotation.components,
        boost_velocity=BMS_inv.boost_velocity,
    )

    abd_interp = abd.interpolate(abd_check.t)

    assert np.allclose(
        np.array(
            [abd_interp.sigma, abd_interp.psi4, abd_interp.psi3, abd_interp.psi2, abd_interp.psi1, abd_interp.psi0]
        ),
        np.array([abd_check.sigma, abd_check.psi4, abd_check.psi3, abd_check.psi2, abd_check.psi1, abd_check.psi0]),
    )


def test_BMS_abd_composition():
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(-100, 100, num=500)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)

    S1 = np.array([1, 2 + 4j, 3, -2 + 4j, 7 - 5j, -3 - 2j, 4, 3 - 2j, 7 + 5j]) * 1e-3
    q1 = np.quaternion(1, 2, 3, 4).normalized()
    v1 = np.array([1, 2, 3]) * 1e-4

    S2 = np.array([-3, 1 - 2j, 5, -1 - 2j, -6 - 4j, 0 + 1j, 3, 0 + 1j, -6 + 4j]) * 1e-3
    q2 = np.quaternion(5, -6, 7, -8).normalized()
    v2 = np.array([-4, 5, -6]) * 1e-4

    BMS1 = bms_transformations.BMSTransformation(
        supertranslation=S1,
        frame_rotation=q1.components,
        boost_velocity=v1,
        order=["supertranslation", "frame_rotation", "boost_velocity"],
    )

    BMS2 = bms_transformations.BMSTransformation(
        supertranslation=S2,
        frame_rotation=q2.components,
        boost_velocity=v2,
        order=["supertranslation", "frame_rotation", "boost_velocity"],
    )

    BMS_composed = BMS2 * BMS1

    abd1 = abd.transform(
        supertranslation=BMS1.supertranslation,
        frame_rotation=BMS1.frame_rotation.components,
        boost_velocity=BMS1.boost_velocity,
    )

    abd2 = abd1.transform(
        supertranslation=BMS2.supertranslation,
        frame_rotation=BMS2.frame_rotation.components,
        boost_velocity=BMS2.boost_velocity,
    )

    abd_composed = abd.transform(
        supertranslation=BMS_composed.supertranslation,
        frame_rotation=BMS_composed.frame_rotation.components,
        boost_velocity=BMS_composed.boost_velocity,
    )

    abd2_interp = abd2.interpolate(
        abd.t[
            np.argmin(abs(abd.t - max(abd2.t[0], abd_composed.t[0]))) : np.argmin(
                abs(abd.t - min(abd2.t[-1], abd_composed.t[-1]))
            )
            + 1
        ]
    )
    abd_composed_interp = abd_composed.interpolate(
        abd.t[
            np.argmin(abs(abd.t - max(abd2.t[0], abd_composed.t[0]))) : np.argmin(
                abs(abd.t - min(abd2.t[-1], abd_composed.t[-1]))
            )
            + 1
        ]
    )

    assert np.allclose(
        np.array(
            [
                abd2_interp.sigma,
                abd2_interp.psi4,
                abd2_interp.psi3,
                abd2_interp.psi2,
                abd2_interp.psi1,
                abd2_interp.psi0,
            ]
        ),
        np.array(
            [
                abd_composed_interp.sigma,
                abd_composed_interp.psi4,
                abd_composed_interp.psi3,
                abd_composed_interp.psi2,
                abd_composed_interp.psi1,
                abd_composed_interp.psi0,
            ]
        ),
    )



import sys
import os


def parse(message):
    if "#prerelease" in message:
        return 'prerelease'

    for pre in ['pre', '']:
        for level in ['patch', 'minor', 'major']:
            if f'#{pre}{level}' in message:
                return f'{pre}{level}'

    return 'patch'


message = os.environ['github_event_head_commit_message']
print(parse(message))



# Copyright (c) 2021, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>
"""Module for operating on gravitational waveforms in various forms

Classes
-------
WaveformBase : Base class
    This is probably not needed directly; it is just used for inheritance by other objects.
WaveformModes: Complex spin-weighted spherical-harmonic modes
    The modes must include all `m` values for a range of `ell` values.  This is the "classic" version of a WaveformBase
    object we might normally think of.
WaveformGrid: Complex quantity evaluated along world lines of grid points on the sphere
    To perform translations or boosts, we need to transform to physical space, along a series of selected world lines
    distributed evenly across the sphere.  These values may need to be interpolated to new time values, and they will
    presumably need to be transformed back to `WaveformModes`.
WaveformInDetector: Real quantities as observed in an inertial detector
    Detectors only measure one polarization, so they deal with real quantities.  Also, data is measured in evenly
    spaced time steps.  This object can be created from a `WaveformModes` object.
WaveformInDetectorFT: (Complex) Fourier transform of a `WaveformInDetector`
    This contains only the positive-frequency values since the transformed data is real.

"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

import sys
import functools
import numba

__version__ = importlib_metadata.version(__name__)

jit = functools.partial(numba.njit, cache=True)
jitclass = numba.experimental.jitclass


def version_info():
    """Show version information about this module and various dependencies"""
    import spherical_functions
    import quaternion
    import scipy
    import numba
    import numpy

    versions = "\n".join(
        [
            f"scri.__version__ = {__version__}",
            f"spherical_functions.__version__ = {spherical_functions.__version__}",
            f"quaternion.__version__ = {quaternion.__version__}",
            f"scipy.__version__ = {scipy.__version__}",
            f"numba.__version__ = {numba.__version__}",
            f"numpy.__version__ = {numpy.__version__}",
        ]
    )
    return versions


# The speed of light is, of course, defined to be exact:
speed_of_light = 299792458.0  # m/s

# The value of the solar mass parameter G*M_sun is known to higher accuracy than either of its factors.  The value
# here is taken from the publication "2015 Selected Astronomical Constants", which can be found at
# <http://asa.usno.navy.mil/SecK/Constants.html>.  This is (one year more current than, but numerically the same as)
# the source cited by the Particle Data Group.  It is given as 1.32712440041e20 m^3/s^2 in the TDB (Barycentric
# Dynamical Time) time scale, which seems to be the more relevant one, and looks like the more standard one for LIGO.
# Dividing by the speed of light squared, we get the mass of the sun in meters; dividing again, we get the mass of
# the sun in seconds:
m_sun_in_meters = 1476.62503851  # m
m_sun_in_seconds = 4.92549094916e-06  # s

# By "IAU 2012 Resolution B2", the astronomical unit is defined to be exactly 1 au = 149597870700 m.  The parsec
# is, in turn, defined as "The distance at which 1 au subtends 1 arc sec: 1 au divided by pi/648000."  Thus, the
# future-proof value of the parsec in meters is
parsec_in_meters = 3.0856775814913672789139379577965e16  # m

FrameType = [UnknownFrameType, Inertial, Coprecessing, Coorbital, Corotating] = range(5)
FrameNames = ["UnknownFrameType", "Inertial", "Coprecessing", "Coorbital", "Corotating"]

DataType = [UnknownDataType, psi0, psi1, psi2, psi3, psi4, sigma, h, hdot, news, psin] = range(11)
DataNames = ["UnknownDataType", "Psi0", "Psi1", "Psi2", "Psi3", "Psi4", "sigma", "h", "hdot", "news", "psin"]
SpinWeights = [sys.maxsize, 2, 1, 0, -1, -2, 2, -2, -2, -2, sys.maxsize]
ConformalWeights = [sys.maxsize, 2, 1, 0, -1, -2, 1, 0, -1, -1, -3]
RScaling = [sys.maxsize, 5, 4, 3, 2, 1, 2, 1, 1, 1, 0]
MScaling = [sys.maxsize, 2, 2, 2, 2, 2, 0, 0, 1, 1, 2]
DataNamesLaTeX = [
    r"\mathrm{unknown data type}",
    r"\psi_0",
    r"\psi_1",
    r"\psi_2",
    r"\psi_3",
    r"\psi_4",
    r"\sigma",
    r"h",
    r"\dot{h}",
    r"\mathrm{n}",
    r"\psi_n",
]
# It might also be worth noting that:
# - the radius `r` has spin weight 0 and boost weight -1
# - a time-derivative `d/du` has spin weight 0 and boost weight -1
# - \eth has spin weight +1; \bar{\eth} has spin weight -1
# - \eth in the GHP formalism has boost weight 0
# - \eth in the original NP formalism has undefined boost weight
# - It seems like `M` should have boost weight 1, but I'll have to think about the implications

# Set up the WaveformModes object, by adding some methods
from .waveform_modes import WaveformModes
from .mode_calculations import (
    LdtVector,
    LVector,
    LLComparisonMatrix,
    LLMatrix,
    LLDominantEigenvector,
    angular_velocity,
    corotating_frame,
    inner_product,
)

from .flux import energy_flux, momentum_flux, angular_momentum_flux, poincare_fluxes, boost_flux


WaveformModes.LdtVector = LdtVector
WaveformModes.LVector = LVector
WaveformModes.LLComparisonMatrix = LLComparisonMatrix
WaveformModes.LLMatrix = LLMatrix
WaveformModes.LLDominantEigenvector = LLDominantEigenvector
WaveformModes.angular_velocity = angular_velocity
from .rotations import (
    rotate_decomposition_basis,
    rotate_physical_system,
    to_coprecessing_frame,
    to_corotating_frame,
    to_inertial_frame,
    align_decomposition_frame_to_modes,
)

WaveformModes.rotate_decomposition_basis = rotate_decomposition_basis
WaveformModes.rotate_physical_system = rotate_physical_system
WaveformModes.to_coprecessing_frame = to_coprecessing_frame
WaveformModes.to_corotating_frame = to_corotating_frame
WaveformModes.to_inertial_frame = to_inertial_frame
WaveformModes.align_decomposition_frame_to_modes = align_decomposition_frame_to_modes
WaveformModes.energy_flux = energy_flux
WaveformModes.momentum_flux = momentum_flux
WaveformModes.angular_momentum_flux = angular_momentum_flux
WaveformModes.boost_flux = boost_flux
WaveformModes.poincare_fluxes = poincare_fluxes


from .waveform_grid import WaveformGrid

# from .waveform_in_detector import WaveformInDetector
from .extrapolation import extrapolate

from .modes_time_series import ModesTimeSeries
from .asymptotic_bondi_data import AsymptoticBondiData

from . import sample_waveforms, SpEC, LVC, utilities
from .SpEC import rpxmb
from .SpEC.file_io import create_abd_from_h5

from .bms_transformations import LorentzTransformation, BMSTransformation

__all__ = [
    "WaveformModes",
    "WaveformGrid",
    "WaveformInDetector",
    "FrameType",
    "UnknownFrameType",
    "Inertial",
    "Coprecessing",
    "Coorbital",
    "Corotating",
    "FrameNames",
    "DataType",
    "UnknownDataType",
    "psi0",
    "psi1",
    "psi2",
    "psi3",
    "psi4",
    "sigma",
    "h",
    "hdot",
    "news",
    "psin",
    "DataNames",
    "DataNamesLaTeX",
    "SpinWeights",
    "ConformalWeights",
    "RScaling",
    "MScaling",
    "speed_of_light",
    "m_sun_in_meters",
    "m_sun_in_seconds",
    "parsec_in_meters",
]



import numpy as np

import sxs
import scri
import spinsfast
import spherical_functions as sf
from spherical_functions import LM_index

import quaternion
from quaternion.calculus import derivative
from quaternion.calculus import indefinite_integral as integrate

from scipy.interpolate import CubicSpline

from sxs.waveforms.alignment import align2d

from functools import partial

from scri.asymptotic_bondi_data.map_to_superrest_frame import time_translation, rotation, MT_to_WM, WM_to_MT

def rel_err_between_abds(abd1, abd2, t1, t2):
    t_array = abd1.t[np.argmin(abs(abd1.t - t1)) : np.argmin(abs(abd1.t - t2)) + 1]

    abd1_interp = abd1.interpolate(t_array)
    abd2_interp = abd2.interpolate(t_array)

    abd_diff = abd1_interp.copy()
    abd_diff.sigma -= abd2_interp.sigma
    abd_diff.psi0 -= abd2_interp.psi0
    abd_diff.psi1 -= abd2_interp.psi1
    abd_diff.psi2 -= abd2_interp.psi2
    abd_diff.psi3 -= abd2_interp.psi3
    abd_diff.psi4 -= abd2_interp.psi4

    rel_err = 0
    rel_err += integrate(abd_diff.sigma.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.sigma.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err += integrate(abd_diff.psi0.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.psi0.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err += integrate(abd_diff.psi1.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.psi1.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err += integrate(abd_diff.psi2.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.psi2.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err += integrate(abd_diff.psi3.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.psi3.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err += integrate(abd_diff.psi4.norm(), abd_diff.t)[-1] / (
        integrate(abd1_interp.psi4.norm(), abd1_interp.t)[-1] or (abd1_interp.t[-1] - abd1_interp.t[0])
    )
    rel_err /= 6

    return rel_err


def map_to_abd_frame(
    self,
    target_abd,
    t_0=0,
    padding_time=250,
    N_itr_maxes={
        "abd": 2,
        "superrest": 2,
        "CoM_transformation": 10,
        "rotation": 10,
        "supertranslation": 10,
    },
    rel_err_tols={"CoM_transformation": 1e-12, "rotation": 1e-12, "supertranslation": 1e-12},
    order=["supertranslation", "rotation", "CoM_transformation"],
    ell_max=None,
    alpha_ell_max=None,
    fix_time_phase_freedom=True,
    nprocs=4,
    print_conv=False,
):
    """Transform an abd object to a target abd object using data at t=0.

    This computes the transformations necessary to map an abd object to a target abd object.
    It uses the function map_to_bms_frame with the target charges computed from the target abd object.

    Parameters
    ----------
    target_abd : AsymptoticBondiData
        Target AsymptoticBondiData to map to.
    t_0: float, optional
        When to map to the target BMS frame.
        Default is 0.
    padding_time : float, optional
        Amount by which to pad around t=0 to speed up computations, i.e.,
        distance from t=0 in each direction to be included in self.interpolate(...).
        This also determines the range over which certain BMS charges will be computed.
        Default is 250.
    N_itr_maxes : dict, optional
        Maximum number of iterations to perform for each transformation.
        For 'abd' , this is the number of iterations to use for the abd procedure.
        For 'superrest', this is the number of iterations to use for the superrest procedure.
        For the other options, these are the number of iterations to use for finding each individual transformation.
        Default is
        N_itr_maxes = {
            'abd':                2,
            'superrest':          2,
            'CoM_transformation': 10,
            'rotation':           10,
            'supertranslation':   10,
        }.
    rel_err_tols : dict, optional
        Relative error tolerances for each transformation.
        Default is
        rel_err_tols = {
            'CoM_transformation': 1e-12,
            'rotation':           1e-12,
            'supertranslation':   1e-12
        }.
    order : list, optional
        Order in which to solve for the BMS transformations.
        Default is ["supertranslation", "rotation", "CoM_transformation"].
    ell_max : int, optional
        Maximum ell to use for SWSH/Grid transformations.
        Default is self.ell_max.
    alpha_ell_max : int, optional
        Maximum ell of the supertranslation to use.
        Default is self.ell_max.
    fix_time_phase_freedom : bool, optional
        Whether or not to fix the time and phase freedom using a 2d minimization scheme.
        Default is True.
    nprocs : int, optional
        Number of cpus to use during parallelization for fixing the time and phase freedom.
        Default is 4. 'None' corresponds to the maximum number. '-1' corresponds to no parallelization.
    print_conv: bool, defaults to False
        Whether or not to print the termination criterion. Default is False.

    Returns
    -------
    abd_prime : AsymptoticBondiData
        Result of self.transform(...) where the input transformations are
        the transformations found in the BMSTransformations object.
    transformations : BMSTransformation
        BMS transformation to map to the target BMS frame.

    """
    abd = self.copy()

    target_strain = scri.asymptotic_bondi_data.map_to_superrest_frame.MT_to_WM(
        2.0 * target_abd.sigma.bar, dataType=scri.h
    )

    time_translation = scri.bms_transformations.BMSTransformation()
    if fix_time_phase_freedom:
        # ensure that they are reasonable close
        energy = abd.bondi_four_momentum()[:, 0]
        target_energy = target_abd.bondi_four_momentum()[:, 0]
        time_translation = scri.bms_transformations.BMSTransformation(
            supertranslation=[
                sf.constant_as_ell_0_mode(
                    abd.t[np.argmin(abs(energy - target_energy[np.argmin(abs(target_abd.t - t_0))]))] - t_0
                )
            ]
        )

    BMS_transformation = (time_translation * scri.bms_transformations.BMSTransformation()).reorder(
        ["supertranslation", "frame_rotation", "boost_velocity"]
    )

    abd_interp = abd.interpolate(
        abd.t[
            np.argmin(abs(abd.t - (t_0 - 1.5 * padding_time))) : np.argmin(abs(abd.t - (t_0 + 1.5 * padding_time))) + 1
        ]
    )

    target_abd_superrest, transformation2, rel_err2 = target_abd.map_to_superrest_frame(
        t_0=t_0,
        padding_time=padding_time,
        N_itr_maxes=N_itr_maxes,
        rel_err_tols=rel_err_tols,
        ell_max=ell_max,
        alpha_ell_max=alpha_ell_max,
        print_conv=print_conv,
        order=order,
    )

    target_strain_superrest = scri.asymptotic_bondi_data.map_to_superrest_frame.MT_to_WM(
        2.0 * target_abd_superrest.sigma.bar, dataType=scri.h
    )

    itr = 0
    rel_err = np.inf
    rel_errs = [np.inf]
    while itr < N_itr_maxes["abd"]:
        if itr == 0:
            abd_interp_prime = abd_interp.transform(
                supertranslation=BMS_transformation.supertranslation,
                frame_rotation=BMS_transformation.frame_rotation.components,
                boost_velocity=BMS_transformation.boost_velocity,
            )

        # find the transformations that map to the superrest frame
        abd_interp_superrest, transformation1, rel_err1 = abd_interp_prime.map_to_superrest_frame(
            t_0=t_0,
            padding_time=padding_time,
            N_itr_maxes=N_itr_maxes,
            rel_err_tols=rel_err_tols,
            ell_max=ell_max,
            alpha_ell_max=alpha_ell_max,
            print_conv=print_conv,
            order=order,
        )

        if fix_time_phase_freedom:
            # 2d align now that we're in the superrest frame
            strain_interp_superrest = scri.asymptotic_bondi_data.map_to_superrest_frame.MT_to_WM(
                2.0 * abd_interp_superrest.sigma.bar, dataType=scri.h
            )

            rel_err, _, res = align2d(
                MT_to_WM(WM_to_MT(strain_interp_superrest), sxs_version=True),
                MT_to_WM(WM_to_MT(target_strain_superrest), sxs_version=True),
                t_0 - padding_time,
                t_0 + padding_time,
                n_brute_force_δt=None,
                n_brute_force_δϕ=None,
                include_modes=None,
                nprocs=nprocs,
            )

            time_phase_transformation = scri.bms_transformations.BMSTransformation(
                supertranslation=[sf.constant_as_ell_0_mode(res.x[0])],
                frame_rotation=quaternion.from_rotation_vector(res.x[1] * np.array([0, 0, 1])).components,
            )
        else:
            time_phase_transformation = scri.bms_transformations.BMSTransformation()

        # compose these transformations in the right order
        BMS_transformation = (
            transformation2.inverse() * (time_phase_transformation * (transformation1 * BMS_transformation))
        ).reorder(["supertranslation", "frame_rotation", "boost_velocity"])

        # obtain the transformed abd object
        abd_interp_prime = abd_interp.transform(
            supertranslation=BMS_transformation.supertranslation,
            frame_rotation=BMS_transformation.frame_rotation.components,
            boost_velocity=BMS_transformation.boost_velocity,
        )

        if fix_time_phase_freedom:
            # find the time/phase transformations
            strain_interp_prime = scri.asymptotic_bondi_data.map_to_superrest_frame.MT_to_WM(
                2.0 * abd_interp_prime.sigma.bar, dataType=scri.h
            )

            rel_err, _, res = align2d(
                MT_to_WM(WM_to_MT(strain_interp_prime), sxs_version=True),
                MT_to_WM(WM_to_MT(target_strain), sxs_version=True),
                t_0 - padding_time,
                t_0 + padding_time,
                n_brute_force_δt=None,
                n_brute_force_δϕ=None,
                include_modes=None,
                nprocs=nprocs,
            )

            time_phase_transformation = scri.bms_transformations.BMSTransformation(
                supertranslation=[sf.constant_as_ell_0_mode(res.x[0])],
                frame_rotation=quaternion.from_rotation_vector(res.x[1] * np.array([0, 0, 1])).components,
            )

            BMS_transformation = (time_phase_transformation * BMS_transformation).reorder(
                ["supertranslation", "frame_rotation", "boost_velocity"]
            )

            # obtain the transformed abd object
            abd_interp_prime = abd_interp.transform(
                supertranslation=BMS_transformation.supertranslation,
                frame_rotation=BMS_transformation.frame_rotation.components,
                boost_velocity=BMS_transformation.boost_velocity,
            )
        else:
            rel_err = rel_err_between_abds(target_abd, abd_interp_prime, t_0 - padding_time, t_0 + padding_time)

        if rel_err < min(rel_errs):
            best_BMS_transformation = BMS_transformation.copy()
            best_rel_err = rel_err
        rel_errs.append(rel_err)

        itr += 1

    if print_conv:
        if not itr < N_itr_maxes["abd"]:
            print(f"BMS: maximum number of iterations reached; the min error was {best_rel_err}.")
        else:
            print(f"BMS: tolerance achieved in {itr} iterations!")

    abd_prime = abd.transform(
        supertranslation=best_BMS_transformation.supertranslation,
        frame_rotation=best_BMS_transformation.frame_rotation.components,
        boost_velocity=best_BMS_transformation.boost_velocity,
    )

    return abd_prime, best_BMS_transformation, best_rel_err



import math
import numpy as np
import quaternion
import spinsfast
import spherical_functions as sf
import scri
import pytest

from conftest import kerr_schild, real_supertranslation

ABD = scri.AsymptoticBondiData
abd = scri.asymptotic_bondi_data


def test_abd_schwarzschild():
    mass = 0.789
    ell_max = 8
    u = np.linspace(0, 100, num=5000)
    psi2, psi1, psi0 = kerr_schild(mass, 0.0, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2)
    expected_four_momentum = np.array([mass, 0, 0, 0])
    computed_four_momentum = abd.bondi_four_momentum()
    expected_angular_momentum = np.zeros(3)
    computed_angular_momentum = abd.bondi_angular_momentum()
    # print()
    # print(f"Expected four-momentum: {expected_four_momentum}")
    # print(f"Computed four-momentum: {computed_four_momentum}")
    assert np.allclose(computed_four_momentum, expected_four_momentum, atol=1e-14, rtol=1e-14)
    assert np.allclose(computed_angular_momentum, expected_angular_momentum, atol=1e-14, rtol=1e-14)
    # print('Bondi-violation norms', abd.bondi_violation_norms)


def test_abd_conformal_factors():
    tolerance = 4e-14
    ell_max = 32
    n_theta = 2 * ell_max + 1
    n_phi = n_theta
    cf = abd.transformations.conformal_factors

    def cf2(v, distorted_grid_rotors):
        from math import sin, cos

        # κ = 1 / [γ(1-v⋅r)]
        # ð(κ⁻¹) = - κ⁻² ðκ
        # ðκ = - κ² ð(κ⁻¹)
        theta_phi = sf.theta_phi(n_theta, n_phi)
        kinv = γ * np.array(
            [
                [1 - v[0] * sin(θ) * cos(ϕ) - v[1] * math.sin(θ) * math.sin(ϕ) - v[2] * math.cos(θ) for θ, ϕ in row]
                for row in theta_phi
            ]
        )
        k = 1 / kinv
        κinv = spinsfast.map2salm(kinv, 0, ell_max)
        κ = spinsfast.map2salm(k, 0, ell_max)
        SWSHs = sf.SWSH_grid(distorted_grid_rotors, 0, ell_max)
        one_over_k2 = np.tensordot(κinv, SWSHs, axes=([-1], [-1]))
        one_over_k_cubed2 = one_over_k2**3
        k2 = 1 / one_over_k2
        ðκ = sf.eth_GHP(κ, 0)
        SWSHs = sf.SWSH_grid(distorted_grid_rotors, 1, ell_max)
        ðk = np.tensordot(ðκ, SWSHs, axes=([-1], [-1]))
        ðk_over_k2 = ðk / k2
        return (
            k2[np.newaxis, :, :],
            ðk_over_k2[np.newaxis, :, :],
            one_over_k2[np.newaxis, :, :],
            one_over_k_cubed2[np.newaxis, :, :],
        )

    frame_rotation = quaternion.one
    boost_velocity = np.array([0.01, 0.02, 0.03])
    β = np.linalg.norm(boost_velocity)
    γ = 1 / math.sqrt(1 - β**2)
    distorted_grid_rotors = abd.transformations.boosted_grid(frame_rotation, boost_velocity, n_theta, n_phi)
    k, ðk_over_k, one_over_k, one_over_k_cubed = cf(boost_velocity, distorted_grid_rotors)
    k2, ðk_over_k2, one_over_k2, one_over_k_cubed2 = cf2(boost_velocity, distorted_grid_rotors)
    assert k.shape == k2.shape
    assert ðk_over_k.shape == ðk_over_k2.shape
    assert one_over_k.shape == one_over_k2.shape
    assert one_over_k_cubed.shape == one_over_k_cubed2.shape
    assert isinstance(k, sf.Grid)
    assert isinstance(ðk_over_k, sf.Grid)
    assert isinstance(one_over_k, sf.Grid)
    assert isinstance(one_over_k_cubed, sf.Grid)
    assert k.s == 0
    assert ðk_over_k.s == 1
    assert one_over_k.s == 0
    assert one_over_k_cubed.s == 0
    assert np.allclose(k.view(np.ndarray), k2, atol=tolerance, rtol=tolerance)
    assert np.allclose(ðk_over_k.view(np.ndarray), ðk_over_k2, atol=tolerance, rtol=tolerance)
    assert np.allclose(one_over_k.view(np.ndarray), one_over_k2, atol=tolerance, rtol=tolerance)
    assert np.allclose(one_over_k_cubed.view(np.ndarray), one_over_k_cubed2, atol=tolerance, rtol=tolerance)


def test_abd_schwarzschild_transform():
    tolerance = 1e-14
    for v in [np.array([0.1, 0.0, 0.0]), np.array([0.0, 0.1, 0.0]), np.array([0.0, 0.0, 0.1])]:
        mass = 1.0
        ell_max = 8
        u = np.linspace(0, 100, num=5000)
        psi2, psi1, psi0 = kerr_schild(mass, 0.0, ell_max)
        abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2)
        rest_mass = abd.bondi_rest_mass()
        β = np.linalg.norm(v)
        γ = 1 / np.sqrt(1 - β**2)
        abdprime = abd.transform(boost_velocity=v)
        transformed_rest_mass = abdprime.bondi_rest_mass()
        transformed_four_momentum = abdprime.bondi_four_momentum()
        expected_four_momentum = mass * γ * np.array([1, *-v])
        # print()
        # print(f"v={v}, β={β}, γ={γ}, βγ={β*γ}")
        # print(f"Expected four-momentum:\n{expected_four_momentum}")
        # print(f"New four-momentum:\n{transformed_four_momentum}")
        assert np.allclose(expected_four_momentum, transformed_four_momentum, atol=tolerance, rtol=tolerance)
        assert np.allclose(rest_mass[0], transformed_rest_mass, atol=tolerance, rtol=tolerance)


def test_abd_bondi_angular_momentum():
    tolerance = 1e-14
    for v in [
        np.array([0.1, 0.0, 0.0]),
        np.array([0.0, 0.1, 0.0]),
        np.array([0.1, 0.1, 0.1]),
        np.array([0.0, 0.0, 0.1]),
    ]:
        mass = 1.0
        spin = 0.456
        ell_max = 8
        u = np.linspace(0, 100, num=5000)
        psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
        abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)
        β = np.linalg.norm(v)
        γ = 1 / np.sqrt(1 - β**2)
        abdprime = abd.transform(boost_velocity=v)
        transformed_angular_momentum = abdprime.bondi_angular_momentum()
        angular_momentum = abd.bondi_angular_momentum()[0]
        expected_angular_momentum = γ * angular_momentum + (1 - γ) * np.dot(angular_momentum, v / β) * (v / β)
        assert np.allclose(expected_angular_momentum, transformed_angular_momentum, atol=tolerance, rtol=tolerance)


def test_abd_kerr():
    v = np.array([0.085, -0.034, 0.1])
    mass = 2.0
    spin = 0.456
    ell_max = 8
    u = np.linspace(0, 100, num=5000)
    psi2, psi1, psi0 = kerr_schild(mass, spin, ell_max)
    abd = ABD.from_initial_values(u, ell_max=ell_max, psi2=psi2, psi1=psi1)
    angular_momentum = abd.bondi_angular_momentum()
    S = abd.bondi_dimensionless_spin()
    tolerance = 1e-14
    # This is true because we are in the center-of-momentum frame
    assert np.allclose(S * mass**2, angular_momentum, atol=tolerance, rtol=tolerance)
    abdprime = abd.transform(boost_velocity=v)
    S_prime = abdprime.bondi_dimensionless_spin()
    tolerance = 1e-14
    assert np.allclose(S_prime[-1], S[-1], atol=tolerance, rtol=tolerance)
    N = abdprime.bondi_boost_charge()
    G = abdprime.bondi_CoM_charge()
    P = abdprime.bondi_four_momentum()
    assert np.allclose(N, G - abdprime.t[:, np.newaxis] * P[:, 1:], atol=tolerance, rtol=tolerance)


def test_abd_WaveformModes():
    # Compare a general BMS transformation of a ~random ABD object to the equivalent transformation
    # of the strain of a WaveformModes object

    tolerance = 4e-12 # should pass with 9e-13; bumped for safety; would be 1e-14 if not for boost
    np.random.seed(123)

    def abd_to_h_data(abd):
        # h = 2 σ̄
        return 2 * abd.sigma.bar.ndarray.copy()

    # Generate initial ABD from Kerr-Schild plus some quadratic-in-time shear data
    mass = 1.23
    spin = 0.35
    ell_max = 6
    u = np.linspace(-10, 10, num=1_000)
    ψ2, ψ1, ψ0 = kerr_schild(mass, spin, ell_max)
    σ0 = np.squeeze(0.01000 * (np.random.rand((ell_max+1)**2, 2).view(complex) - (0.5+0.5j)))
    σ0[:sf.LM_index(2,-2,0)] = 0
    σ̇0 = np.squeeze(0.00020 * (np.random.rand((ell_max+1)**2, 2).view(complex) - (0.5+0.5j)))
    σ̇0[:sf.LM_index(2,-2,0)] = 0
    σ̈0 = np.squeeze(0.00003 * (np.random.rand((ell_max+1)**2, 2).view(complex) - (0.5+0.5j)))
    σ̈0[:sf.LM_index(2,-2,0)] = 0
    abd = ABD.from_initial_values(u, ell_max=ell_max, sigma0=σ0, sigmadot0=σ̇0, sigmaddot0=σ̈0, psi2=ψ2, psi1=ψ1, psi0=ψ0)
    h = scri.WaveformModes(
        t=abd.t, data=abd_to_h_data(abd), ell_min=abd.ell_min, ell_max=abd.ell_max,
        frameType=scri.Inertial, dataType=scri.h, r_is_scaled_out=True, m_is_scaled_out=True
    )

    # Generate a random supertranslation
    α = real_supertranslation(np.squeeze(0.01 * (np.random.rand((ell_max+1)**2, 2).view(complex) - (0.5+0.5j))))
    R = np.random.randn(4)  # Will be normalized automatically
    v⃗ = 0.01 * (np.random.rand(3)-0.5)

    # Transform ABD and comparable WaveformModes objects
    abdprime = abd.transform(supertranslation=α, frame_rotation=R, boost_velocity=v⃗)
    hprime = h.transform(supertranslation=α, frame_rotation=R, boost_velocity=v⃗)

    # Interpolate to same set of times
    t = scri.extrapolation.intersection(hprime.t, abdprime.t)
    hprime = hprime.interpolate(t)
    abdprime = abdprime.interpolate(t)

    assert np.allclose(
        hprime.data,
        abd_to_h_data(abdprime)[:, 4:],
        atol=tolerance, rtol=tolerance
    )



# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import numpy as np
import quaternion
from numpy import *
import pytest
import scri

from conftest import random_waveform


def test_parity_projections():
    for dataType in [scri.psi0, scri.psi1, scri.psi2, scri.psi3, scri.psi4, scri.h]:
        w = random_waveform(dataType=dataType)
        ## x
        x = w.x_parity_symmetric_part
        assert np.array_equal(x.data, x.x_parity_conjugate.data)
        assert np.array_equal(x.data, x.x_parity_symmetric_part.data)
        assert np.array_equal(np.zeros_like(x.data), x.x_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(x.t), x.x_parity_violation_squared)
        x = w.x_parity_antisymmetric_part
        assert np.array_equal(x.data, -x.x_parity_conjugate.data)
        assert np.array_equal(x.data, x.x_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(x.data), x.x_parity_symmetric_part.data)
        assert np.array_equal(x.norm(), x.x_parity_violation_squared)
        ## y
        y = w.y_parity_symmetric_part
        assert np.array_equal(y.data, y.y_parity_conjugate.data)
        assert np.array_equal(y.data, y.y_parity_symmetric_part.data)
        assert np.array_equal(np.zeros_like(y.data), y.y_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(y.t), y.y_parity_violation_squared)
        y = w.y_parity_antisymmetric_part
        assert np.array_equal(y.data, -y.y_parity_conjugate.data)
        assert np.array_equal(y.data, y.y_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(y.data), y.y_parity_symmetric_part.data)
        assert np.array_equal(y.norm(), y.y_parity_violation_squared)
        ## z
        z = w.z_parity_symmetric_part
        assert np.array_equal(z.data, z.z_parity_conjugate.data)
        assert np.array_equal(z.data, z.z_parity_symmetric_part.data)
        assert np.array_equal(np.zeros_like(z.data), z.z_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(z.t), z.z_parity_violation_squared)
        z = w.z_parity_antisymmetric_part
        assert np.array_equal(z.data, -z.z_parity_conjugate.data)
        assert np.array_equal(z.data, z.z_parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(z.data), z.z_parity_symmetric_part.data)
        assert np.array_equal(z.norm(), z.z_parity_violation_squared)
        ## xyz
        xyz = w.parity_symmetric_part
        assert np.array_equal(xyz.data, xyz.parity_conjugate.data)
        assert np.array_equal(xyz.data, xyz.parity_symmetric_part.data)
        assert np.array_equal(np.zeros_like(xyz.data), xyz.parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(xyz.t), xyz.parity_violation_squared)
        xyz = w.parity_antisymmetric_part
        assert np.array_equal(xyz.data, -xyz.parity_conjugate.data)
        assert np.array_equal(xyz.data, xyz.parity_antisymmetric_part.data)
        assert np.array_equal(np.zeros_like(xyz.data), xyz.parity_symmetric_part.data)
        assert np.array_equal(xyz.norm(), xyz.parity_violation_squared)



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import copy
import math
import numpy as np
import spherical_functions as sf
import scri.sample_waveforms as samples
import spinsfast
import pytest

# from conftest import Rs

slow = pytest.mark.slow


def test_time_translation():
    """Test pure time translation, in both time_translation and supertranslation forms"""
    dt = 1.469
    alpha00 = math.sqrt(4 * math.pi) * dt
    w1 = samples.constant_waveform()
    w2 = w1.transform(time_translation=dt)
    w3 = w1.transform(supertranslation=[alpha00])
    assert np.allclose(w1.t, w2.t + dt, rtol=0.0, atol=2e-15)
    assert np.allclose(w1.data, w2.data, rtol=0.0, atol=4e-14)
    assert np.allclose(w2.t, w3.t, rtol=0.0, atol=0.0)
    assert np.allclose(w2.data, w3.data, rtol=0.0, atol=0.0)


def test_BMS_rotation(Rs):
    """Compare full BMS transformation machinery to simple Wigner-D rotation"""
    w1 = samples.constant_waveform()
    for i, R in enumerate(Rs):
        w2 = w1.copy()
        w3 = w1.copy()
        w2.rotate_decomposition_basis(R)
        w3 = w3.transform(frame_rotation=R.components)
        assert np.allclose(w2.data, w3.data, rtol=1e-15, atol=4e-13)


@slow
def test_space_translation():
    """Compare code-transformed waveform to analytically transformed waveform"""
    print("")
    ell_max = 8
    for s in range(-2, 2 + 1):
        for ell in range(abs(s), ell_max + 1):
            print("\tWorking on spin s =", s, ", ell =", ell)
            for m in range(-ell, ell + 1):
                for space_translation in [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]:
                    auxiliary_waveforms = {}
                    for i in range(s + 2):
                        auxiliary_waveforms[f"psi{4-i}_modes"] = samples.single_mode_proportional_to_time(s=i - 2)
                        auxiliary_waveforms[f"psi{4-i}_modes"].data *= 0
                    w_m1 = samples.single_mode_proportional_to_time(s=s, ell=ell, m=m).transform(
                        space_translation=space_translation,
                        **auxiliary_waveforms,
                    )
                    w_m2 = samples.single_mode_proportional_to_time_supertranslated(
                        s=s, ell=ell, m=m, space_translation=np.array(space_translation)
                    )
                    i1A = np.argmin(abs(w_m1.t - (w_m1.t[0] + 2 * np.linalg.norm(space_translation))))
                    i1B = np.argmin(abs(w_m1.t - (w_m1.t[-1] - 2 * np.linalg.norm(space_translation))))
                    i2A = np.argmin(abs(w_m2.t - w_m1.t[i1A]))
                    i2B = np.argmin(abs(w_m2.t - w_m1.t[i1B]))
                    assert np.allclose(w_m1.t[i1A : i1B + 1], w_m2.t[i2A : i2B + 1], rtol=0.0, atol=1e-16), (
                        w_m1.t[i1A],
                        w_m2.t[i2A],
                        w_m1.t[i1B],
                        w_m2.t[i2B],
                        w_m1.t[i1A : i1B + 1].shape,
                        w_m2.t[i2A : i2B + 1].shape,
                    )
                    data1 = w_m1.data[i1A : i1B + 1]
                    data2 = w_m2.data[i2A : i2B + 1]
                    assert np.allclose(data1, data2, rtol=0.0, atol=5e-14), (
                        [s, ell, m],
                        space_translation,
                        [
                            abs(data1 - data2).max(),
                            data1.ravel()[np.argmax(abs(data1 - data2))],
                            data2.ravel()[np.argmax(abs(data1 - data2))],
                        ],
                        [
                            np.unravel_index(np.argmax(abs(data1 - data2)), data1.shape)[0],
                            list(
                                sf.LM_range(abs(s), ell_max)[
                                    np.unravel_index(np.argmax(abs(data1 - data2)), data1.shape)[1]
                                ]
                            ),
                        ],
                    )


@slow
def test_hyper_translation():
    """Compare code-transformed waveform to analytically transformed waveform"""
    print("")
    ell_max = 4
    for s in range(-2, 2 + 1):
        for ell in range(abs(s), ell_max + 1):
            for m in range(-ell, ell + 1):
                print("\tWorking on spin s =", s, ", ell =", ell, ", m =", m)
                for ellpp, mpp in sf.LM_range(2, ell_max):
                    supertranslation = np.zeros((sf.LM_total_size(0, ell_max),), dtype=complex)
                    if mpp == 0:
                        supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0
                    elif mpp < 0:
                        supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0
                        supertranslation[sf.LM_index(ellpp, -mpp, 0)] = (-1.0) ** mpp
                    elif mpp > 0:
                        supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0j
                        supertranslation[sf.LM_index(ellpp, -mpp, 0)] = (-1.0) ** mpp * -1.0j
                    max_displacement = abs(
                        spinsfast.salm2map(supertranslation, 0, ell_max, 4 * ell_max + 1, 4 * ell_max + 1)
                    ).max()
                    auxiliary_waveforms = {}
                    for i in range(s + 2):
                        auxiliary_waveforms[f"psi{4-i}_modes"] = samples.single_mode_proportional_to_time(s=i - 2)
                        auxiliary_waveforms[f"psi{4-i}_modes"].data *= 0
                    w_m1 = samples.single_mode_proportional_to_time(s=s, ell=ell, m=m).transform(
                        supertranslation=supertranslation,
                        **auxiliary_waveforms,
                    )
                    w_m2 = samples.single_mode_proportional_to_time_supertranslated(
                        s=s, ell=ell, m=m, supertranslation=supertranslation
                    )
                    i1A = np.argmin(abs(w_m1.t - (w_m1.t[0] + 2 * max_displacement)))
                    i1B = np.argmin(abs(w_m1.t - (w_m1.t[-1] - 2 * max_displacement)))
                    i2A = np.argmin(abs(w_m2.t - w_m1.t[i1A]))
                    i2B = np.argmin(abs(w_m2.t - w_m1.t[i1B]))
                    assert np.allclose(w_m1.t[i1A : i1B + 1], w_m2.t[i2A : i2B + 1], rtol=0.0, atol=1e-16), (
                        w_m1.t[i1A],
                        w_m2.t[i2A],
                        w_m1.t[i1B],
                        w_m2.t[i2B],
                        w_m1.t[i1A : i1B + 1].shape,
                        w_m2.t[i2A : i2B + 1].shape,
                    )
                    data1 = w_m1.data[i1A : i1B + 1]
                    data2 = w_m2.data[i2A : i2B + 1]
                    assert np.allclose(data1, data2, rtol=0.0, atol=5e-14), (
                        [s, ell, m],
                        supertranslation,
                        [
                            abs(data1 - data2).max(),
                            data1.ravel()[np.argmax(abs(data1 - data2))],
                            data2.ravel()[np.argmax(abs(data1 - data2))],
                        ],
                        [
                            np.unravel_index(np.argmax(abs(data1 - data2)), data1.shape)[0],
                            list(
                                sf.LM_range(abs(s), ell_max)[
                                    np.unravel_index(np.argmax(abs(data1 - data2)), data1.shape)[1]
                                ]
                            ),
                        ],
                    )


def test_supertranslation_inverses():
    w1 = samples.random_waveform_proportional_to_time(rotating=False)
    ell_max = 4
    for ellpp, mpp in sf.LM_range(0, ell_max):
        supertranslation = np.zeros((sf.LM_total_size(0, ell_max),), dtype=complex)
        if mpp == 0:
            supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0
        elif mpp < 0:
            supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0
            supertranslation[sf.LM_index(ellpp, -mpp, 0)] = (-1.0) ** mpp
        elif mpp > 0:
            supertranslation[sf.LM_index(ellpp, mpp, 0)] = 1.0j
            supertranslation[sf.LM_index(ellpp, -mpp, 0)] = (-1.0) ** mpp * -1.0j
        max_displacement = abs(spinsfast.salm2map(supertranslation, 0, ell_max, 4 * ell_max + 1, 4 * ell_max + 1)).max()
        w2 = copy.deepcopy(w1)
        w2 = w2.transform(supertranslation=supertranslation)
        w2 = w2.transform(supertranslation=-supertranslation)

        i1A = np.argmin(abs(w1.t - (w1.t[0] + 3 * max_displacement)))
        i1B = np.argmin(abs(w1.t - (w1.t[-1] - 3 * max_displacement)))
        i2A = np.argmin(abs(w2.t - w1.t[i1A]))
        i2B = np.argmin(abs(w2.t - w1.t[i1B]))
        try:
            assert np.allclose(w1.t[i1A : i1B + 1], w2.t[i2A : i2B + 1], rtol=0.0, atol=1e-15), (
                w1.t[i1A],
                w2.t[i2A],
                w1.t[i1B],
                w2.t[i2B],
                w1.t[i1A : i1B + 1].shape,
                w2.t[i2A : i2B + 1].shape,
            )
        except ValueError:
            print("Indices:\n\t", i1A, i1B, i2A, i2B)
            print("Times:\n\t", w1.t[i1A], w1.t[i1B], w2.t[i2A], w2.t[i2B])
            raise
        data1 = w1.data[i1A : i1B + 1]
        data2 = w2.data[i2A : i2B + 1]
        try:
            assert np.allclose(data1, data2, rtol=5e-10, atol=5e-14), [
                abs(data1 - data2).max(),
                data1.ravel()[np.argmax(abs(data1 - data2))],
                data2.ravel()[np.argmax(abs(data1 - data2))],
                np.unravel_index(np.argmax(abs(data1 - data2)), data1.shape),
            ]
            # list(sf.LM_range(0, ell_max)[np.unravel_index(np.argmax(abs(data1-data2)),
            #                                               data1.shape)[1]])])
        except:
            print("Indices:\n\t", i1A, i1B, i2A, i2B)
            print("Times:\n\t", w1.t[i1A], w1.t[i1B], w2.t[i2A], w2.t[i2B])
            raise


def test_boost_inverses():
    """Boost a waveform, invert that boost, and check that the result is identical"""
    # We'll try an easier case with boost 1e-2, then a harder case with boost 1e-1.  For the latter, we need to
    # increase the ell_max substantially.
    for beta, ell_max in [(1e-2, 8), (1e-1, 14)]:
        for v in [np.array([0.0, 0.0, beta]), np.array([0.0, beta, 0.0]), np.array([beta, 0.0, 0.0])]:
            # Construct a simple sample waveform.  This fills up all the modes with reasonable values, for a good test.
            space_translation = np.array([0.1, 0.0, 0.0])
            w1 = samples.single_mode_constant_rotation(s=-2, ell=2, m=2, omega=0.3, t_0=-10.0, t_1=10.0, dt=1.0 / 200.0)
            w1 = w1.transform(space_translation=space_translation)

            # Don't include mass scaling, because that's more subtle
            w1.m_is_scaled_out = False

            # Make a copy, transform, and transform back
            w2 = w1.copy()
            w2 = w2.transform(
                boost_velocity=v, n_theta=2 * (ell_max + 1) + 1, n_phi=2 * (ell_max + 1) + 1, ell_max=ell_max
            )
            w2 = w2.transform(boost_velocity=-v, ell_max=w1.ell_max)

            # Interpolate the original onto the new time set, because we've lost some time steps
            w1 = w1.interpolate(w2.t)

            # Check that everything agrees
            # print(abs(w1.data-w2.data).max())
            assert w1._allclose(w2, atol=4e-14, rtol=0)



import math
import tempfile
import pathlib
import numpy as np
import h5py
import scri
import pytest


def generate_bit_widths(bit_width):
    possible_widths = 2 ** np.arange(0, int(np.log2(bit_width)))
    bit_widths = []
    while np.sum(bit_widths) < bit_width:
        next_width = np.random.choice(possible_widths)
        if np.sum(bit_widths) + next_width <= bit_width:
            bit_widths.append(next_width)
    return tuple(bit_widths)


@pytest.mark.parametrize("bit_width", [8, 16, 32, 64])
def test_multishuffle_reversibility(bit_width):
    dt = np.dtype(f"u{bit_width//8}")
    np.random.seed(123)
    data = np.random.randint(0, high=2 ** bit_width, size=5_000, dtype=dt)
    for bit_widths in [(1,) * bit_width, (8,) * (bit_width // 8)] + [generate_bit_widths(bit_width) for _ in range(10)]:
        shuffle = scri.utilities.multishuffle(bit_widths)
        unshuffle = scri.utilities.multishuffle(bit_widths, forward=False)
        assert np.array_equal(data, unshuffle(shuffle(data))), bit_widths


@pytest.mark.parametrize("bit_width", [8, 16, 32, 64])
def test_multishuffle_like_hdf5(bit_width):
    dt = np.dtype(f"u{bit_width//8}")
    np.random.seed(1234)
    data = np.random.randint(0, high=2 ** bit_width, size=5_000, dtype=dt)

    # Save the data to file via h5py, then extract the raw data to see what
    # HDF5's shuffle looks like
    with tempfile.TemporaryDirectory() as temp_dir:
        file_name = pathlib.Path(temp_dir) / "test.h5"
        with h5py.File(file_name, "w") as f:
            f.create_dataset("data", data=data, shuffle=True, chunks=(data.size,))
        with h5py.File(file_name, "r") as f:
            ds = f["data"]
            filter_mask, raw_data_bytes = ds.id.read_direct_chunk((0,))
            hdf5_raw_data = np.frombuffer(raw_data_bytes, dtype=dt)

    # Shuffle with our function
    shuffle = scri.utilities.multishuffle((8,) * (bit_width // 8))
    scri_shuffle_data = shuffle(data)

    # Check that they are equal
    assert np.array_equal(scri_shuffle_data, hdf5_raw_data)



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

import pytest
import numpy as np
from numpy import *
import quaternion
import spherical_functions as sf
import scri

from conftest import linear_waveform, constant_waveform, random_waveform, delta_waveform


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_identity_rotation(w):
    # Rotation by 1 should be identity operation
    W_in = w()
    W_out = w()
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    W_out.rotate_decomposition_basis(quaternion.one)
    assert W_out.ensure_validity(alter=False)
    assert np.array_equal(W_out.t, W_in.t)
    assert np.array_equal(W_out.frame, W_in.frame)
    assert np.array_equal(W_out.data, W_in.data)
    assert np.array_equal(W_out.LM, W_in.LM)
    assert W_out.ell_min == W_in.ell_min
    assert W_out.ell_max == W_in.ell_max
    for h_in, h_out in zip(W_in.history, W_out.history[:-1]):
        assert h_in == h_out.replace(
            f"{type(W_out).__name__}_{str(W_out.num)}", f"{type(W_in).__name__}_{str(W_in.num)}"
        ) or (h_in.startswith("# ") and h_out.startswith("# "))
    assert W_out.frameType == W_in.frameType
    assert W_out.dataType == W_in.dataType
    assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
    assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
    assert isinstance(W_out.num, int)
    assert W_out.num != W_in.num


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_rotation_invariants(w):
    # A random rotation should leave everything but data and frame the
    # same (except num, of course)
    W_in = w()
    W_out = w()
    np.random.seed(hash("test_rotation_invariants") % 4294967294)  # Use mod to get in an acceptable range
    W_out.rotate_decomposition_basis(np.quaternion(*np.random.uniform(-1, 1, 4)).normalized())
    assert W_in.ensure_validity(alter=False)
    assert W_out.ensure_validity(alter=False)
    assert np.array_equal(W_out.t, W_in.t)
    assert not np.array_equal(W_out.frame, W_in.frame)  # This SHOULD change
    assert not np.array_equal(W_out.data, W_in.data)  # This SHOULD change
    assert W_out.ell_min == W_in.ell_min
    assert W_out.ell_max == W_in.ell_max
    assert np.array_equal(W_out.LM, W_in.LM)
    for h_in, h_out in zip(W_in.history[:-3], W_out.history[:-5]):
        assert h_in == h_out.replace(
            f"{type(W_out).__name__}_{str(W_out.num)}", f"{type(W_in).__name__}_{str(W_in.num)}"
        ) or (h_in.startswith("# ") and h_out.startswith("# "))
    assert W_out.frameType == W_in.frameType
    assert W_out.dataType == W_in.dataType
    assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
    assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
    assert W_out.num != W_in.num


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_constant_versus_series(w):
    # A random rotation should leave everything but data and frame the
    # same (except num, of course)
    W_const = w()
    W_series = w()
    np.random.seed(hash("test_constant_versus_series") % 4294967294)  # Use mod to get in an acceptable range
    W_const.rotate_decomposition_basis(np.quaternion(*np.random.uniform(-1, 1, 4)).normalized())
    W_series.rotate_decomposition_basis(
        np.array([np.quaternion(*np.random.uniform(-1, 1, 4)).normalized()] * W_series.n_times)
    )
    assert W_const.ensure_validity(alter=False)
    assert W_series.ensure_validity(alter=False)
    assert np.array_equal(W_series.t, W_const.t)
    assert not np.array_equal(W_series.frame, W_const.frame)  # This SHOULD change
    assert not np.array_equal(W_series.data, W_const.data)  # This SHOULD change
    assert W_series.ell_min == W_const.ell_min
    assert W_series.ell_max == W_const.ell_max
    assert np.array_equal(W_series.LM, W_const.LM)
    for h_const, h_series in zip(W_const.history[:-5], W_series.history[:-11]):
        assert h_const == h_series.replace(
            f"{type(W_series).__name__}_{str(W_series.num)}", f"{type(W_const).__name__}_{str(W_const.num)}"
        ) or (h_const.startswith("# ") and h_series.startswith("# "))
    assert W_series.frameType == W_const.frameType
    assert W_series.dataType == W_const.dataType
    assert W_series.r_is_scaled_out == W_const.r_is_scaled_out
    assert W_series.m_is_scaled_out == W_const.m_is_scaled_out
    assert W_series.num != W_const.num


@pytest.mark.parametrize("w", [linear_waveform, constant_waveform, random_waveform])
def test_rotation_inversion(w):
    # Rotation followed by the inverse rotation should leave
    # everything the same (except that the frame data will be either a
    # 1 or a series of 1s)
    np.random.seed(hash("test_rotation_inversion") % 4294967294)  # Use mod to get in an acceptable range
    W_in = w()
    assert W_in.ensure_validity(alter=False)
    # We loop over (1) a single constant rotation, and (2) an array of random rotations
    for R_basis in [
        np.quaternion(*np.random.uniform(-1, 1, 4)).normalized(),
        np.array([np.quaternion(*np.random.uniform(-1, 1, 4)).normalized()] * W_in.n_times),
    ]:
        W_out = w()
        W_out.rotate_decomposition_basis(R_basis)
        W_out.rotate_decomposition_basis(~R_basis)
        assert W_out.ensure_validity(alter=False)
        assert np.array_equal(W_out.t, W_in.t)
        assert np.max(np.abs(W_out.frame - W_in.frame)) < 1e-15
        assert np.allclose(W_out.data, W_in.data, atol=W_in.ell_max ** 4 ** 4e-14, rtol=W_in.ell_max ** 4 * 4e-14)
        assert W_out.ell_min == W_in.ell_min
        assert W_out.ell_max == W_in.ell_max
        assert np.array_equal(W_out.LM, W_in.LM)
        for h_in, h_out in zip(W_in.history[:-3], W_out.history[:-5]):
            assert h_in == h_out.replace(
                f"{type(W_out).__name__}_{str(W_out.num)}", f"{type(W_in).__name__}_{str(W_in.num)}"
            ) or (h_in.startswith("# datetime") and h_out.startswith("# datetime"))
        assert W_out.frameType == W_in.frameType
        assert W_out.dataType == W_in.dataType
        assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
        assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
        assert W_out.num != W_in.num


def test_rotations_of_0_0_mode(Rs):
    # The (ell,m)=(0,0) mode should be rotationally invariant
    n_copies = 10
    W_in = delta_waveform(0, 0, begin=-10.0, end=100.0, n_times=n_copies * len(Rs), ell_min=0, ell_max=8)
    assert W_in.ensure_validity(alter=False)
    W_out = scri.WaveformModes(W_in)
    R_basis = np.array([R for R in Rs for i in range(n_copies)])
    W_out.rotate_decomposition_basis(R_basis)
    assert W_out.ensure_validity(alter=False)
    assert np.array_equal(W_out.t, W_in.t)
    assert np.max(np.abs(W_out.frame - R_basis)) == 0.0
    assert np.array_equal(W_out.data, W_in.data)
    assert W_out.ell_min == W_in.ell_min
    assert W_out.ell_max == W_in.ell_max
    assert np.array_equal(W_out.LM, W_in.LM)
    for h_in, h_out in zip(W_in.history, W_out.history[:-1]):
        assert h_in == h_out.replace(
            f"{type(W_out).__name__}_{str(W_out.num)}", f"{type(W_in).__name__}_{str(W_in.num)}"
        ) or (h_in.startswith("# ") and h_out.startswith("# "))
    assert W_out.frameType == W_in.frameType
    assert W_out.dataType == W_in.dataType
    assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
    assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
    assert W_out.num != W_in.num


def test_rotations_of_each_mode_individually(Rs):
    ell_min = 0
    ell_max = 8  # sf.ell_max is just too much; this test is too slow, and ell=8 should be fine
    R_basis = Rs
    Ds = np.empty((len(Rs), sf.LMpM_total_size(ell_min, ell_max)), dtype=complex)
    for i, R in enumerate(Rs):
        Ds[i, :] = sf.Wigner_D_matrices(R, ell_min, ell_max)
    for ell in range(ell_max + 1):
        first_zeros = np.zeros((len(Rs), sf.LM_total_size(ell_min, ell - 1)), dtype=complex)
        later_zeros = np.zeros((len(Rs), sf.LM_total_size(ell + 1, ell_max)), dtype=complex)
        for Mp in range(-ell, ell):
            W_in = delta_waveform(ell, Mp, begin=-10.0, end=100.0, n_times=len(Rs), ell_min=ell_min, ell_max=ell_max)
            # Now, the modes are f^{\ell,m[} = \delta^{\ell,mp}_{L,Mp}
            assert W_in.ensure_validity(alter=False)
            W_out = scri.WaveformModes(W_in)
            W_out.rotate_decomposition_basis(Rs)
            assert W_out.ensure_validity(alter=False)
            assert np.array_equal(W_out.t, W_in.t)
            assert np.max(np.abs(W_out.frame - R_basis)) == 0.0
            i_D0 = sf.LMpM_index(ell, Mp, -ell, ell_min)
            assert np.array_equal(W_out.data[:, : sf.LM_total_size(ell_min, ell - 1)], first_zeros)
            if ell < ell_max:
                assert np.array_equal(
                    W_out.data[:, sf.LM_total_size(ell_min, ell - 1) : -sf.LM_total_size(ell + 1, ell_max)],
                    Ds[:, i_D0 : i_D0 + (2 * ell + 1)],
                )
                assert np.array_equal(W_out.data[:, -sf.LM_total_size(ell + 1, ell_max) :], later_zeros)
            else:
                assert np.array_equal(
                    W_out.data[:, sf.LM_total_size(ell_min, ell - 1) :], Ds[:, i_D0 : i_D0 + (2 * ell + 1)]
                )
            assert W_out.ell_min == W_in.ell_min
            assert W_out.ell_max == W_in.ell_max
            assert np.array_equal(W_out.LM, W_in.LM)
            for h_in, h_out in zip(W_in.history, W_out.history[:-1]):
                assert h_in == h_out.replace(type(W_out).__name__ + str(W_out.num), type(W_in).__name__ + str(W_in.num))
            assert W_out.frameType == W_in.frameType
            assert W_out.dataType == W_in.dataType
            assert W_out.r_is_scaled_out == W_in.r_is_scaled_out
            assert W_out.m_is_scaled_out == W_in.m_is_scaled_out
            assert W_out.num != W_in.num
